@@ -32,6 +32,12 @@
  * (https://github.com/edenhill/librdkafka)
  */
 
+#ifdef _MSC_VER
+#define  _CRT_SECURE_NO_WARNINGS /* Silence nonsense on MSVC */
+#endif
+
+#include "../src/rd.h"
+
 #define _GNU_SOURCE /* for strndup() */
 #include <ctype.h>
 #include <signal.h>
@@ -46,6 +52,12 @@
 #include "rd.h"
 #include "rdtime.h"
 
+#ifdef _MSC_VER
+#include "../win32/wingetopt.h"
+#include "../win32/wintime.h"
+#endif
+
+
 static int run = 1;
 static int forever = 1;
 static rd_ts_t dispintvl = 1000;
@@ -106,10 +118,9 @@ static struct {
         rd_ts_t  latency_sum;
         int      latency_cnt;
         int64_t  last_offset;
-} cnt = {};
+} cnt;
 
 
-/* Returns wall clock time in microseconds */
 uint64_t wall_clock (void) {
         struct timeval tv;
         gettimeofday(&tv, NULL);
@@ -147,6 +158,20 @@ static void offset_commit_cb (rd_kafka_t *rk, rd_kafka_resp_err_t err,
         }
 }
 
+/**
+ * @brief Add latency measurement
+ */
+static void latency_add (int64_t ts, const char *who) {
+        if (ts > cnt.latency_hi)
+                cnt.latency_hi = ts;
+        if (!cnt.latency_lo || ts < cnt.latency_lo)
+                cnt.latency_lo = ts;
+        cnt.latency_last = ts;
+        cnt.latency_cnt++;
+        cnt.latency_sum += ts;
+        if (latency_fp)
+                fprintf(latency_fp, "%"PRIu64"\n", ts);
+}
 
 
 static void msg_delivered (rd_kafka_t *rk,
@@ -166,6 +191,14 @@ static void msg_delivered (rd_kafka_t *rk,
                 cnt.bytes_dr_ok += rkmessage->len;
         }
 
+        if (latency_mode) {
+                /* Extract latency */
+                int64_t source_ts;
+                if (sscanf(rkmessage->payload, "LATENCY:%"SCNd64,
+                           &source_ts) == 1)
+                        latency_add(wall_clock() - source_ts, "producer");
+        }
+
 
 	if ((rkmessage->err &&
 	     (cnt.msgs_dr_err < 50 ||
@@ -269,15 +302,7 @@ static void msg_consume (rd_kafka_message_t *rkmessage, void *opaque) {
                            &remote_ts) == 1) {
                         ts = wall_clock() - remote_ts;
                         if (ts > 0 && ts < (1000000 * 60 * 5)) {
-                                if (ts > cnt.latency_hi)
-                                        cnt.latency_hi = ts;
-                                if (!cnt.latency_lo || ts < cnt.latency_lo)
-                                        cnt.latency_lo = ts;
-                                cnt.latency_last = ts;
-                                cnt.latency_cnt++;
-                                cnt.latency_sum += ts;
-				if (latency_fp)
-					fprintf(latency_fp, "%"PRIu64"\n", ts);
+                                latency_add(ts, "consumer");
                         } else {
                                 if (verbosity >= 1)
                                         printf("Received latency timestamp is too far off: %"PRId64"us (message offset %"PRId64"): ignored\n",
@@ -335,8 +360,8 @@ static uint64_t json_parse_fields (const char *json, const char **end,
                                    const char *field1, const char *field2) {
         const char *t = json;
         const char *t2;
-        int len1 = strlen(field1);
-        int len2 = strlen(field2);
+        int len1 = (int)strlen(field1);
+        int len2 = (int)strlen(field2);
 
         while ((t2 = strstr(t, field1))) {
                 uint64_t v;
@@ -369,16 +394,16 @@ static uint64_t json_parse_fields (const char *json, const char **end,
  */
 static void json_parse_stats (const char *json) {
         const char *t;
-        const int max_avgs = 100; /* max number of brokers to scan for rtt */
-        uint64_t avg_rtt[max_avgs+1];
+#define MAX_AVGS 100 /* max number of brokers to scan for rtt */
+        uint64_t avg_rtt[MAX_AVGS+1];
         int avg_rtt_i     = 0;
 
         /* Store totals at end of array */
-        avg_rtt[max_avgs]     = 0;
+        avg_rtt[MAX_AVGS]     = 0;
 
         /* Extract all broker RTTs */
         t = json;
-        while (avg_rtt_i < max_avgs && *t) {
+        while (avg_rtt_i < MAX_AVGS && *t) {
                 avg_rtt[avg_rtt_i] = json_parse_fields(t, &t,
                                                        "\"rtt\":",
                                                        "\"avg\":");
@@ -388,14 +413,14 @@ static void json_parse_stats (const char *json) {
                         continue;
 
 
-                avg_rtt[max_avgs] += avg_rtt[avg_rtt_i];
+                avg_rtt[MAX_AVGS] += avg_rtt[avg_rtt_i];
                 avg_rtt_i++;
         }
 
         if (avg_rtt_i > 0)
-                avg_rtt[max_avgs] /= avg_rtt_i;
+                avg_rtt[MAX_AVGS] /= avg_rtt_i;
 
-        cnt.avg_rtt = avg_rtt[max_avgs];
+        cnt.avg_rtt = avg_rtt[MAX_AVGS];
 }
 
 
@@ -419,16 +444,11 @@ static void print_stats (rd_kafka_t *rk,
 	rd_ts_t t_total;
         static int rows_written = 0;
         int print_header;
+        double latency_avg = 0.0f;
         char extra[512];
         int extra_of = 0;
         *extra = '\0';
 
-#define EXTRA_PRINTF(fmt...) do {                                       \
-                if (extra_of < sizeof(extra))                           \
-                        extra_of += snprintf(extra+extra_of,            \
-                                             sizeof(extra)-extra_of, fmt); \
-        } while (0)
-
 	if (!(otype & _OTYPE_FORCE) &&
             (((otype & _OTYPE_SUMMARY) && verbosity == 0) ||
              cnt.t_last + dispintvl > now))
@@ -445,6 +465,10 @@ static void print_stats (rd_kafka_t *rk,
 	else
 		t_total = 1;
 
+        if (latency_mode && cnt.latency_cnt)
+                latency_avg = (double)cnt.latency_sum /
+                        (double)cnt.latency_cnt;
+
         if (mode == 'P') {
 
                 if (otype & _OTYPE_TAB) {
@@ -472,6 +496,13 @@ static void print_stats (rd_kafka_t *rk,
                                 COL_HDR("outq");
                                 if (report_offset)
                                         COL_HDR("offset");
+                                if (latency_mode) {
+                                        COL_HDR("lat_curr");
+                                        COL_HDR("lat_avg");
+                                        COL_HDR("lat_lo");
+                                        COL_HDR("lat_hi");
+                                }
+
                                 ROW_END();
                         }
 
@@ -491,6 +522,12 @@ static void print_stats (rd_kafka_t *rk,
                                  rk ? (uint64_t)rd_kafka_outq_len(rk) : 0);
                         if (report_offset)
                                 COL_PR64("offset", (uint64_t)cnt.last_offset);
+                        if (latency_mode) {
+                                COL_PRF("lat_curr", cnt.latency_last / 1000.0f);
+                                COL_PRF("lat_avg", latency_avg / 1000.0f);
+                                COL_PRF("lat_lo", cnt.latency_lo / 1000.0f);
+                                COL_PRF("lat_hi", cnt.latency_hi / 1000.0f);
+                        }
                         ROW_END();
                 }
 
@@ -500,7 +537,7 @@ static void print_stats (rd_kafka_t *rk,
                                "%"PRIu64" delivered "
                                "(offset %"PRId64", %"PRIu64" failed) "
                                "in %"PRIu64"ms: %"PRIu64" msgs/s and "
-                               "%.02f Mb/s, "
+                               "%.02f MB/s, "
                                "%"PRIu64" produce failures, %i in queue, "
                                "%s compression\n",
                                cnt.msgs, cnt.bytes,
@@ -514,11 +551,6 @@ static void print_stats (rd_kafka_t *rk,
                 }
 
         } else {
-                float latency_avg = 0.0f;
-
-                if (latency_mode && cnt.latency_cnt)
-                        latency_avg = (double)cnt.latency_sum /
-                                (float)cnt.latency_cnt;
 
                 if (otype & _OTYPE_TAB) {
                         if (print_header) {
@@ -564,7 +596,7 @@ static void print_stats (rd_kafka_t *rk,
 
                 if (otype & _OTYPE_SUMMARY) {
                         if (latency_avg >= 1.0f)
-                                extra_of += snprintf(extra+extra_of,
+                                extra_of += rd_snprintf(extra+extra_of,
                                                      sizeof(extra)-extra_of,
                                                      ", latency "
                                                      "curr/avg/lo/hi "
@@ -576,7 +608,7 @@ static void print_stats (rd_kafka_t *rk,
 ;
                         printf("%% %"PRIu64" messages (%"PRIu64" bytes) "
                                "consumed in %"PRIu64"ms: %"PRIu64" msgs/s "
-                               "(%.02f Mb/s)"
+                               "(%.02f MB/s)"
                                "%s\n",
                                cnt.msgs, cnt.bytes,
                                t_total / 1000,
@@ -593,7 +625,7 @@ static void print_stats (rd_kafka_t *rk,
                         printf("%% INTERVAL: %"PRIu64" messages "
                                "(%"PRIu64" bytes) "
                                "consumed in %"PRIu64"ms: %"PRIu64" msgs/s "
-                               "(%.02f Mb/s)"
+                               "(%.02f MB/s)"
                                "%s\n",
                                i_msgs, i_bytes,
                                i_time / 1000,
@@ -614,6 +646,76 @@ static void sig_usr1 (int sig) {
 	rd_kafka_dump(stdout, global_rk);
 }
 
+
+/**
+ * @brief Read config from file
+ * @returns -1 on error, else 0.
+ */
+static int read_conf_file (rd_kafka_conf_t *conf,
+                           rd_kafka_topic_conf_t *tconf, const char *path) {
+        FILE *fp;
+        char buf[512];
+        int line = 0;
+        char errstr[512];
+
+        if (!(fp = fopen(path, "r"))) {
+                fprintf(stderr, "%% Failed to open %s: %s\n",
+                        path, strerror(errno));
+                return -1;
+        }
+
+        while (fgets(buf, sizeof(buf), fp)) {
+                char *s = buf;
+                char *t;
+                rd_kafka_conf_res_t r = RD_KAFKA_CONF_UNKNOWN;
+
+                line++;
+
+                while (isspace((int)*s))
+                        s++;
+
+                if (!*s || *s == '#')
+                        continue;
+
+                if ((t = strchr(buf, '\n')))
+                        *t = '\0';
+
+                t = strchr(buf, '=');
+                if (!t || t == s || !*(t+1)) {
+                        fprintf(stderr, "%% %s:%d: expected key=value\n",
+                                path, line);
+                        fclose(fp);
+                        return -1;
+                }
+
+                *(t++) = '\0';
+
+                /* Try property on topic config first */
+                if (tconf)
+                        r = rd_kafka_topic_conf_set(tconf, s, t,
+                                                    errstr, sizeof(errstr));
+
+                /* Try global config */
+                if (r == RD_KAFKA_CONF_UNKNOWN)
+                        r = rd_kafka_conf_set(conf, s, t,
+                                              errstr, sizeof(errstr));
+
+                if (r == RD_KAFKA_CONF_OK)
+                        continue;
+
+                fprintf(stderr, "%% %s:%d: %s=%s: %s\n",
+                        path, line, s, t, errstr);
+                fclose(fp);
+                return -1;
+        }
+
+        fclose(fp);
+
+        return 0;
+}
+
+
+
 int main (int argc, char **argv) {
 	char *brokers = NULL;
 	char mode = 'C';
@@ -623,12 +725,12 @@ int main (int argc, char **argv) {
 	int opt;
 	int sendflags = 0;
 	char *msgpattern = "librdkafka_performance testing!";
-	int msgsize = strlen(msgpattern);
+	int msgsize = (int)strlen(msgpattern);
 	const char *debug = NULL;
 	rd_ts_t now;
 	char errstr[512];
 	uint64_t seq = 0;
-	int seed = time(NULL);
+	int seed = (int)time(NULL);
         rd_kafka_t *rk;
 	rd_kafka_topic_t *rkt;
 	rd_kafka_conf_t *conf;
@@ -646,6 +748,7 @@ int main (int argc, char **argv) {
         double dtmp;
         int rate_sleep = 0;
 	rd_kafka_topic_partition_list_t *topics;
+        int exitcode = 0;
 
 	/* Kafka configuration */
 	conf = rd_kafka_conf_new();
@@ -653,9 +756,11 @@ int main (int argc, char **argv) {
 	rd_kafka_conf_set_throttle_cb(conf, throttle_cb);
         rd_kafka_conf_set_offset_commit_cb(conf, offset_commit_cb);
 
-	/* Quick termination */
-	snprintf(tmp, sizeof(tmp), "%i", SIGIO);
+#ifdef SIGIO
+        /* Quick termination */
+	rd_snprintf(tmp, sizeof(tmp), "%i", SIGIO);
 	rd_kafka_conf_set(conf, "internal.termination.signal", tmp, NULL, 0);
+#endif
 
 	/* Producer config */
 	rd_kafka_conf_set(conf, "queue.buffering.max.messages", "500000",
@@ -804,6 +909,12 @@ int main (int argc, char **argv) {
 			*val = '\0';
 			val++;
 
+                        if (!strcmp(name, "file")) {
+                                if (read_conf_file(conf, topic_conf, val) == -1)
+                                        exit(1);
+                                break;
+                        }
+
 			res = RD_KAFKA_CONF_UNKNOWN;
 			/* Try "topic." prefixed properties on topic
 			 * conf first, and then fall through to global if
@@ -923,6 +1034,7 @@ int main (int argc, char **argv) {
 			"  -s <size>    Message size (producer)\n"
 			"  -k <key>     Message key (producer)\n"
 			"  -c <cnt>     Messages to transmit/receive\n"
+			"  -x <cnt>     Hard exit after transmitting <cnt> messages (producer)\n"
 			"  -D           Copy/Duplicate data buffer (producer)\n"
 			"  -i <ms>      Display interval\n"
 			"  -m <msg>     Message payload pattern\n"
@@ -944,6 +1056,7 @@ int main (int argc, char **argv) {
 			"will be set on topic object.\n"
 			"               Use '-X list' to see the full list\n"
 			"               of supported properties.\n"
+                        "  -X file=<path> Read config from file.\n"
 			"  -T <intvl>   Enable statistics from librdkafka at "
 			"specified interval (ms)\n"
                         "  -Y <command> Pipe statistics to <command>\n"
@@ -956,6 +1069,7 @@ int main (int argc, char **argv) {
                         "               Needs two matching instances, one\n"
                         "               consumer and one producer, both\n"
                         "               running with the -l switch.\n"
+                        "  -l           Producer: per-message latency stats\n"
 			"  -A <file>    Write per-message latency stats to "
 			"<file>. Requires -l\n"
                         "  -O           Report produced offset (producer)\n"
@@ -983,7 +1097,9 @@ int main (int argc, char **argv) {
                        seed, verbosity);
 	srand(seed);
 	signal(SIGINT, stop);
+#ifdef SIGUSR1
 	signal(SIGUSR1, sig_usr1);
+#endif
 
 
 	if (debug &&
@@ -1001,7 +1117,7 @@ int main (int argc, char **argv) {
         if (!stats_intvlstr) {
                 /* if no user-desired stats, adjust stats interval
                  * to the display interval. */
-                snprintf(tmp, sizeof(tmp), "%"PRId64, dispintvl / 1000);
+                rd_snprintf(tmp, sizeof(tmp), "%"PRId64, dispintvl / 1000);
         }
 
         if (rd_kafka_conf_set(conf, "statistics.interval.ms",
@@ -1018,6 +1134,7 @@ int main (int argc, char **argv) {
         if (stats_intvlstr) {
                 /* User enabled stats (-T) */
 
+#ifndef _MSC_VER
                 if (stats_cmd) {
                         if (!(stats_fp = popen(stats_cmd, "we"))) {
                                 fprintf(stderr,
@@ -1026,6 +1143,7 @@ int main (int argc, char **argv) {
                                 exit(1);
                         }
                 } else
+#endif
                         stats_fp = stdout;
         }
 
@@ -1041,18 +1159,18 @@ int main (int argc, char **argv) {
 		char *sbuf;
 		char *pbuf;
 		int outq;
-		int keylen = key ? strlen(key) : 0;
+		int keylen = key ? (int)strlen(key) : 0;
 		off_t rof = 0;
 		size_t plen = strlen(msgpattern);
 		int partition = partitions ? partitions[0] :
 			RD_KAFKA_PARTITION_UA;
 
                 if (latency_mode) {
-                        msgsize = strlen("LATENCY:") + 
-                                strlen("18446744073709551615 ")+1;
+                        msgsize = (int)(strlen("LATENCY:") +
+                                strlen("18446744073709551615 ")+1);
                         sendflags |= RD_KAFKA_MSG_F_COPY;
 		} else if (do_seq) {
-                        int minlen = strlen("18446744073709551615 ")+1;
+                        int minlen = (int)strlen("18446744073709551615 ")+1;
                         if (msgsize < minlen)
                                 msgsize = minlen;
 
@@ -1066,7 +1184,7 @@ int main (int argc, char **argv) {
 		while (rof < msgsize) {
 			size_t xlen = RD_MIN((size_t)msgsize-rof, plen);
 			memcpy(sbuf+rof, msgpattern, xlen);
-			rof += xlen;
+			rof += (off_t)xlen;
 		}
 
 		if (msgcnt == -1)
@@ -1090,9 +1208,6 @@ int main (int argc, char **argv) {
 
                 global_rk = rk;
 
-		if (debug)
-			rd_kafka_set_log_level(rk, 7);
-
 		/* Add broker(s) */
 		if (brokers && rd_kafka_brokers_add(rk, brokers) < 1) {
 			fprintf(stderr, "%% No valid brokers specified\n");
@@ -1125,10 +1240,10 @@ int main (int argc, char **argv) {
 			}
 
                         if (latency_mode) {
-                                snprintf(sbuf, msgsize-1,
+                                rd_snprintf(sbuf, msgsize-1,
                                          "LATENCY:%"PRIu64,  wall_clock());
                         } else if (do_seq) {
-                                snprintf(sbuf,
+                                rd_snprintf(sbuf,
                                          msgsize-1, "%"PRIu64": ", seq);
                                 seq++;
 			}
@@ -1148,20 +1263,19 @@ int main (int argc, char **argv) {
 			       rd_kafka_produce(rkt, partition,
 						sendflags, pbuf, msgsize,
 						key, keylen, NULL) == -1) {
-				if (errno == ESRCH)
+				rd_kafka_resp_err_t err = rd_kafka_last_error();
+				if (err == RD_KAFKA_RESP_ERR__UNKNOWN_PARTITION)
 					printf("%% No such partition: "
-                                               "%"PRId32"\n", partition);
-				else if ((errno != ENOBUFS && verbosity >= 1) ||
-                                         verbosity >= 3)
+						   "%"PRId32"\n", partition);
+				else if (verbosity >= 3 ||
+					(err != RD_KAFKA_RESP_ERR__QUEUE_FULL && verbosity >= 1))
 					printf("%% produce error: %s%s\n",
-					       rd_kafka_err2str(
-						       rd_kafka_errno2err(
-							       errno)),
-					       errno == ENOBUFS ?
-					       " (backpressure)":"");
+						   rd_kafka_err2str(err),
+						   err == RD_KAFKA_RESP_ERR__QUEUE_FULL ?
+						   " (backpressure)" : "");
 
 				cnt.tx_err++;
-				if (errno != ENOBUFS) {
+				if (err != RD_KAFKA_RESP_ERR__QUEUE_FULL) {
 					run = 0;
 					break;
 				}
@@ -1188,8 +1302,13 @@ int main (int argc, char **argv) {
 			cnt.msgs++;
 			cnt.bytes += msgsize;
 
-                        if (rate_sleep)
+                        if (rate_sleep) {
+#ifdef _MSC_VER
+                                Sleep(rate_sleep / 1000);
+#else
                                 usleep(rate_sleep);
+#endif
+                        }
 
 			/* Must poll to handle delivery reports */
 			rd_kafka_poll(rk, 0);
@@ -1202,8 +1321,6 @@ int main (int argc, char **argv) {
                         printf("%% All messages produced, "
                                "now waiting for %li deliveries\n",
                                msgs_wait_cnt);
-		if (debug)
-			rd_kafka_dump(stdout, rk);
 
 		/* Wait for messages to be delivered */
                 while (run && rd_kafka_poll(rk, 1000) != -1)
@@ -1224,9 +1341,6 @@ int main (int argc, char **argv) {
 			       cnt.tx_err, cnt.tx,
 			       ((double)cnt.tx_err / (double)cnt.tx) * 100.0);
 
-		if (debug)
-			rd_kafka_dump(stdout, rk);
-
 		/* Destroy topic */
 		rd_kafka_topic_destroy(rkt);
 
@@ -1236,6 +1350,8 @@ int main (int argc, char **argv) {
 
 		free(sbuf);
 
+                exitcode = cnt.msgs == cnt.msgs_dr_ok ? 0 : 1;
+
 	} else if (mode == 'C') {
 		/*
 		 * Consumer
@@ -1255,9 +1371,6 @@ int main (int argc, char **argv) {
 
                 global_rk = rk;
 
-		if (debug)
-			rd_kafka_set_log_level(rk, 7);
-
 		/* Add broker(s) */
 		if (brokers && rd_kafka_brokers_add(rk, brokers) < 1) {
 			fprintf(stderr, "%% No valid brokers specified\n");
@@ -1278,9 +1391,8 @@ int main (int argc, char **argv) {
 				partitions[i], start_offset, rkqu);
 
 			if (r == -1) {
-				fprintf(stderr, "%% Error creating queue: %s\n",
-					rd_kafka_err2str(
-						rd_kafka_errno2err(errno)));
+                                fprintf(stderr, "%% Error creating queue: %s\n",
+                                        rd_kafka_err2str(rd_kafka_last_error()));
 				exit(1);
 			}
 		}
@@ -1291,7 +1403,7 @@ int main (int argc, char **argv) {
 			 * an error signaling (if rkmessage->err is set).
 			 */
 			uint64_t fetch_latency;
-			int r;
+			ssize_t r;
 
 			fetch_latency = rd_clock();
 
@@ -1321,10 +1433,9 @@ int main (int argc, char **argv) {
 			}
 
 			cnt.t_fetch_latency += rd_clock() - fetch_latency;
-			if (r == -1)
-				fprintf(stderr, "%% Error: %s\n",
-					rd_kafka_err2str(
-						rd_kafka_errno2err(errno)));
+                        if (r == -1)
+                                fprintf(stderr, "%% Error: %s\n",
+                                        rd_kafka_err2str(rd_kafka_last_error()));
 
 			print_stats(rk, mode, otype, compression);
 
@@ -1335,12 +1446,11 @@ int main (int argc, char **argv) {
 
 		/* Stop consuming */
 		for (i=0 ; i<(size_t)partition_cnt ; ++i) {
-			int r = rd_kafka_consume_stop(rkt, i);
+			int r = rd_kafka_consume_stop(rkt, (int32_t)i);
 			if (r == -1) {
-				fprintf(stderr,
-					"%% Error in consume_stop: %s\n",
-					rd_kafka_err2str(
-						rd_kafka_errno2err(errno)));
+                                fprintf(stderr,
+                                        "%% Error in consume_stop: %s\n",
+                                        rd_kafka_err2str(rd_kafka_last_error()));
 			}
 		}
 		rd_kafka_queue_destroy(rkqu);
@@ -1379,9 +1489,6 @@ int main (int argc, char **argv) {
 
                 global_rk = rk;
 
-		if (debug)
-			rd_kafka_set_log_level(rk, 7);
-
 		/* Add broker(s) */
 		if (brokers && rd_kafka_brokers_add(rk, brokers) < 1) {
 			fprintf(stderr, "%% No valid brokers specified\n");
@@ -1436,7 +1543,9 @@ int main (int argc, char **argv) {
 		fclose(latency_fp);
 
         if (stats_fp) {
+#ifndef _MSC_VER
                 pclose(stats_fp);
+#endif
                 stats_fp = NULL;
         }
 
@@ -1448,5 +1557,5 @@ int main (int argc, char **argv) {
 	/* Let background threads clean up and terminate cleanly. */
 	rd_kafka_wait_destroyed(2000);
 
-	return 0;
+	return exitcode;
 }
