@@ -63,14 +63,30 @@
 #elif defined sun
   #include <sys/byteorder.h>
   #include <sys/isa_defs.h>
+#define __LITTLE_ENDIAN 1234
+#define __BIG_ENDIAN 4321
 #ifdef _BIG_ENDIAN
+#define __BYTE_ORDER __BIG_ENDIAN
 #define be64toh(x) (x)
 #define be32toh(x) (x)
 #define be16toh(x) (x)
+#define le16toh(x) ((uint16_t)BSWAP_16(x))
+#define le32toh(x) BSWAP_32(x)
+#define le64toh(x) BSWAP_64(x)
+#define htole16(x) ((uint16_t)BSWAP_16(x))
+#define htole32(x) BSWAP_32(x)
+#define htole64(x) BSWAP_64(x)
 # else
+#define __BYTE_ORDER __LITTLE_ENDIAN
 #define be64toh(x) BSWAP_64(x)
 #define be32toh(x) ntohl(x)
 #define be16toh(x) ntohs(x)
+#define le16toh(x) (x)
+#define le32toh(x) (x)
+#define le64toh(x) (x)
+#define htole16(x) (x)
+#define htole32(x) (x)
+#define htole64(x) (x)
 #endif /* sun */
 
 #elif defined __APPLE__
@@ -108,8 +124,11 @@
 
 
 
-
-#ifndef be64toh
+/*
+ * On Solaris, be64toh is a function, not a macro, so there's no need to error
+ * if it's not defined.
+ */
+#if !defined(__sun) && !defined(be64toh)
 #error Missing definition for be64toh
 #endif
 
