@@ -113,7 +113,8 @@ struct rd_kafka_broker_s { /* rd_kafka_broker_t */
 					      * See RD_KAFKA_FEATURE_* in
 					      * rdkafka_proto.h */
 
-	struct rd_kafka_ApiVersion *rkb_ApiVersions;     /* Broker's supported APIs.*/
+        struct rd_kafka_ApiVersion *rkb_ApiVersions; /* Broker's supported APIs
+                                                      * (MUST be sorted) */
 	size_t                      rkb_ApiVersions_cnt;
 	rd_interval_t               rkb_ApiVersion_fail_intvl; /* Controls how long
 								* the fallback proto
@@ -137,6 +138,7 @@ struct rd_kafka_broker_s { /* rd_kafka_broker_t */
                                               * and dropped. */
                 rd_atomic64_t zbuf_grow;     /* Compression/decompression buffer grows needed */
                 rd_atomic64_t buf_grow;      /* rkbuf grows needed */
+                rd_atomic64_t wakeups;       /* Poll wakeups */
 	} rkb_c;
 
         int                 rkb_req_timeouts;  /* Current value */
@@ -159,6 +161,7 @@ struct rd_kafka_broker_s { /* rd_kafka_broker_t */
 	rd_kafka_bufq_t     rkb_waitresps;
 	rd_kafka_bufq_t     rkb_retrybufs;
 
+	rd_avg_t            rkb_avg_int_latency;/* Current internal latency period*/
 	rd_avg_t            rkb_avg_rtt;        /* Current RTT period */
 	rd_avg_t            rkb_avg_throttle;   /* Current throttle period */
 
@@ -174,12 +177,18 @@ struct rd_kafka_broker_s { /* rd_kafka_broker_t */
         char               *rkb_logname;
         mtx_t               rkb_logname_lock;
 
-        rd_ts_t             rkb_ts_connect;       /* Last connection attempt */
+        int                 rkb_wakeup_fd[2];     /* Wake-up fds (r/w) to wake
+                                                   * up from IO-wait when
+                                                   * queues have content. */
+        int                 rkb_toppar_wakeup_fd; /* Toppar msgq wakeup fd,
+                                                   * this is rkb_wakeup_fd[1]
+                                                   * if enabled. */
+        rd_interval_t       rkb_connect_intvl;    /* Reconnect throttling */
 
 	rd_kafka_secproto_t rkb_proto;
 
 	int                 rkb_down_reported;    /* Down event reported */
-#if WITH_SASL
+#if WITH_SASL_CYRUS
 	rd_kafka_timer_t    rkb_sasl_kinit_refresh_tmr;
 #endif
 
@@ -195,6 +204,16 @@ struct rd_kafka_broker_s { /* rd_kafka_broker_t */
 #define rd_kafka_broker_unlock(rkb) mtx_unlock(&(rkb)->rkb_lock)
 
 
+/**
+ * @brief Broker comparator
+ */
+static RD_UNUSED RD_INLINE int rd_kafka_broker_cmp (const void *_a,
+                                                    const void *_b) {
+        const rd_kafka_broker_t *a = _a, *b = _b;
+        return (int)(a - b);
+}
+
+
 /**
  * @returns true if broker supports \p features, else false.
  */
@@ -207,6 +226,13 @@ int rd_kafka_broker_supports (rd_kafka_broker_t *rkb, int features) {
 	return r;
 }
 
+int16_t rd_kafka_broker_ApiVersion_supported (rd_kafka_broker_t *rkb,
+                                              int16_t ApiKey,
+                                              int16_t minver, int16_t maxver,
+                                              int *featuresp);
+
+int rd_kafka_broker_get_state (rd_kafka_broker_t *rkb);
+
 rd_kafka_broker_t *rd_kafka_broker_find_by_nodeid (rd_kafka_t *rk,
 						   int32_t nodeid);
 rd_kafka_broker_t *rd_kafka_broker_find_by_nodeid0 (rd_kafka_t *rk,
@@ -237,7 +263,8 @@ rd_kafka_broker_t *rd_kafka_broker_any (rd_kafka_t *rk, int state,
                                                        void *opaque),
                                         void *opaque);
 
-rd_kafka_broker_t *rd_kafka_broker_any_usable (rd_kafka_t *rk, int timeout_ms);
+rd_kafka_broker_t *rd_kafka_broker_any_usable (rd_kafka_t *rk, int timeout_ms,
+                                               int do_lock);
 
 rd_kafka_broker_t *rd_kafka_broker_prefer (rd_kafka_t *rk, int32_t broker_id, int state);
 
@@ -248,14 +275,12 @@ void rd_kafka_broker_fail (rd_kafka_broker_t *rkb,
 			   int level, rd_kafka_resp_err_t err,
 			   const char *fmt, ...);
 
-void rd_kafka_topic_leader_query0 (rd_kafka_t *rk, rd_kafka_itopic_t *rkt,
-				   int do_rk_lock);
-#define rd_kafka_topic_leader_query(rk,rkt) \
-	rd_kafka_topic_leader_query0(rk,rkt,1)
 void rd_kafka_broker_destroy_final (rd_kafka_broker_t *rkb);
-#define rd_kafka_broker_destroy(rkb)                                    \
-        rd_refcnt_destroywrapper(&(rkb)->rkb_refcnt,                    \
-                                 rd_kafka_broker_destroy_final(rkb))
+
+static RD_INLINE RD_UNUSED void rd_kafka_broker_destroy (rd_kafka_broker_t *rkb) {
+        rd_refcnt_destroywrapper(&rkb->rkb_refcnt,
+                                 rd_kafka_broker_destroy_final(rkb));
+}
 
 void rd_kafka_broker_update (rd_kafka_t *rk, rd_kafka_secproto_t proto,
                              const struct rd_kafka_metadata_broker *mdb);
@@ -275,13 +300,11 @@ void rd_kafka_dr_msgq (rd_kafka_itopic_t *rkt,
 		       rd_kafka_msgq_t *rkmq, rd_kafka_resp_err_t err);
 
 void rd_kafka_broker_buf_enq1 (rd_kafka_broker_t *rkb,
-                               int16_t ApiKey,
                                rd_kafka_buf_t *rkbuf,
                                rd_kafka_resp_cb_t *resp_cb,
                                void *opaque);
 
 void rd_kafka_broker_buf_enq_replyq (rd_kafka_broker_t *rkb,
-                                     int16_t ApiKey,
                                      rd_kafka_buf_t *rkbuf,
                                      rd_kafka_replyq_t replyq,
                                      rd_kafka_resp_cb_t *resp_cb,
@@ -289,13 +312,6 @@ void rd_kafka_broker_buf_enq_replyq (rd_kafka_broker_t *rkb,
 
 void rd_kafka_broker_buf_retry (rd_kafka_broker_t *rkb, rd_kafka_buf_t *rkbuf);
 
-void rd_kafka_broker_metadata_req (rd_kafka_broker_t *rkb,
-                                   int all_topics,
-                                   rd_kafka_itopic_t *only_rkt,
-                                   rd_kafka_replyq_t replyq,
-                                   const char *reason);
-void rd_kafka_broker_metadata_req_op (rd_kafka_broker_t *rkb,
-				      rd_kafka_op_t *rko);
 
 rd_kafka_broker_t *rd_kafka_broker_internal (rd_kafka_t *rk);
 
@@ -304,6 +320,7 @@ void msghdr_print (rd_kafka_t *rk,
 		   int hexdump);
 
 const char *rd_kafka_broker_name (rd_kafka_broker_t *rkb);
+void rd_kafka_broker_wakeup (rd_kafka_broker_t *rkb);
 
 int rd_kafka_brokers_get_state_version (rd_kafka_t *rk);
 int rd_kafka_brokers_wait_state_change (rd_kafka_t *rk, int stored_version,
