@@ -35,6 +35,10 @@
 
 #include "rdkafka_int.h"
 #include "rdkafka_feature.h"
+#include "rdkafka_interceptor.h"
+#if WITH_PLUGINS
+#include "rdkafka_plugin.h"
+#endif
 
 struct rd_kafka_property {
 	rd_kafka_conf_scope_t scope;
@@ -50,7 +54,8 @@ struct rd_kafka_property {
 		_RK_C_PTR,  /* Only settable through special set functions */
                 _RK_C_PATLIST, /* Pattern list */
                 _RK_C_KSTR, /* Kafka string */
-		_RK_C_ALIAS /* Alias: points to other property through .sdef */
+                _RK_C_ALIAS, /* Alias: points to other property through .sdef */
+                _RK_C_INTERNAL, /* Internal, don't expose to application */
 	} type;
 	int   offset;
 	const char *desc;
@@ -67,6 +72,20 @@ struct rd_kafka_property {
 	/* Value validator (STR) */
 	int (*validate) (const struct rd_kafka_property *prop,
 			 const char *val, int ival);
+
+        /* Configuration object constructors and destructor for use when
+         * the property value itself is not used, or needs extra care. */
+        void (*ctor) (int scope, void *pconf);
+        void (*dtor) (int scope, void *pconf);
+        void (*copy) (int scope, void *pdst, const void *psrc,
+                      void *dstptr, const void *srcptr,
+                      size_t filter_cnt, const char **filter);
+
+        rd_kafka_conf_res_t (*set) (int scope, void *pconf,
+                                    const char *name, const char *value,
+                                    void *dstptr,
+                                    rd_kafka_conf_set_mode_t set_mode,
+                                    char *errstr, size_t errstr_size);
 };
 
 
@@ -74,6 +93,10 @@ struct rd_kafka_property {
 #define _RKT(field) offsetof(rd_kafka_topic_conf_t, field)
 
 
+static rd_kafka_conf_res_t
+rd_kafka_anyconf_get0 (const void *conf, const struct rd_kafka_property *prop,
+                       char *dest, size_t *dest_size);
+
 
 /**
  * @brief Validate \p broker.version.fallback property.
@@ -105,7 +128,7 @@ static const struct rd_kafka_property rd_kafka_properties[] = {
 	"Indicates the builtin features for this build of librdkafka. "
 	"An application can either query this value or attempt to set it "
 	"with its list of required features to check for library support.",
-	0, 0x7fffffff, 0xff,
+	0, 0x7fffffff, 0xffff,
 	.s2i = {
 #if WITH_ZLIB
 		{ 0x1, "gzip" },
@@ -116,12 +139,18 @@ static const struct rd_kafka_property rd_kafka_properties[] = {
 #if WITH_SSL
 		{ 0x4, "ssl" },
 #endif
-#if WITH_SASL
-		{ 0x8, "sasl" },
-#endif
+                { 0x8, "sasl" },
 		{ 0x10, "regex" },
-#if WITH_LZ4
 		{ 0x20, "lz4" },
+#if defined(_MSC_VER) || WITH_SASL_CYRUS
+                { 0x40, "sasl_gssapi" },
+#endif
+                { 0x80, "sasl_plain" },
+#if WITH_SASL_SCRAM
+                { 0x100, "sasl_scram" },
+#endif
+#if WITH_PLUGINS
+                { 0x200, "plugins" },
 #endif
 		{ 0, NULL }
 		}
@@ -130,7 +159,7 @@ static const struct rd_kafka_property rd_kafka_properties[] = {
 	  "Client identifier.",
 	  .sdef =  "rdkafka" },
 	{ _RK_GLOBAL, "metadata.broker.list", _RK_C_STR, _RK(brokerlist),
-	  "Initial list of brokers. "
+	  "Initial list of brokers as a CSV list of broker host or host:port. "
 	  "The application may also use `rd_kafka_brokers_add()` to add "
 	  "brokers during runtime." },
 	{ _RK_GLOBAL, "bootstrap.servers", _RK_C_ALIAS, 0,
@@ -171,19 +200,22 @@ static const struct rd_kafka_property rd_kafka_properties[] = {
 	  "The metadata is automatically refreshed on error and connect. "
 	  "Use -1 to disable the intervalled refresh.",
 	  -1, 3600*1000, 5*60*1000 },
-	{ _RK_GLOBAL, "metadata.max.age.ms", _RK_C_ALIAS,
-	  .sdef = "topic.metadata.refresh.interval.ms" },
-	{ _RK_GLOBAL, "topic.metadata.refresh.fast.cnt", _RK_C_INT,
-	  _RK(metadata_refresh_fast_cnt),
-	  "When a topic looses its leader this number of metadata requests "
-	  "are sent with `topic.metadata.refresh.fast.interval.ms` interval "
-	  "disregarding the `topic.metadata.refresh.interval.ms` value. "
-	  "This is used to recover quickly from transitioning leader brokers.",
-	  0, 1000, 10 },
-	{ _RK_GLOBAL, "topic.metadata.refresh.fast.interval.ms", _RK_C_INT,
-	  _RK(metadata_refresh_fast_interval_ms),
-	  "See `topic.metadata.refresh.fast.cnt` description",
-	  1, 60*1000, 250 },
+	{ _RK_GLOBAL, "metadata.max.age.ms", _RK_C_INT,
+          _RK(metadata_max_age_ms),
+          "Metadata cache max age. "
+          "Defaults to metadata.refresh.interval.ms * 3",
+          1, 24*3600*1000, -1 },
+        { _RK_GLOBAL, "topic.metadata.refresh.fast.interval.ms", _RK_C_INT,
+          _RK(metadata_refresh_fast_interval_ms),
+          "When a topic loses its leader a new metadata request will be "
+          "enqueued with this initial interval, exponentially increasing "
+          "until the topic metadata has been refreshed. "
+          "This is used to recover quickly from transitioning leader brokers.",
+          1, 60*1000, 250 },
+        { _RK_GLOBAL, "topic.metadata.refresh.fast.cnt", _RK_C_INT,
+          _RK(metadata_refresh_fast_cnt),
+          "*Deprecated: No longer used.*",
+          0, 1000, 10 },
         { _RK_GLOBAL, "topic.metadata.refresh.sparse", _RK_C_BOOL,
           _RK(metadata_refresh_sparse),
           "Sparse metadata requests (consumes less network bandwidth)",
@@ -208,6 +240,8 @@ static const struct rd_kafka_property rd_kafka_properties[] = {
 			{ RD_KAFKA_DBG_SECURITY, "security" },
 			{ RD_KAFKA_DBG_FETCH,    "fetch" },
 			{ RD_KAFKA_DBG_FEATURE,  "feature" },
+                        { RD_KAFKA_DBG_INTERCEPTOR, "interceptor" },
+                        { RD_KAFKA_DBG_PLUGIN,   "plugin" },
 			{ RD_KAFKA_DBG_ALL,      "all" },
 		} },
 	{ _RK_GLOBAL, "socket.timeout.ms", _RK_C_INT, _RK(socket_timeout_ms),
@@ -217,8 +251,8 @@ static const struct rd_kafka_property rd_kafka_properties[] = {
 	  _RK(socket_blocking_max_ms),
 	  "Maximum time a broker socket operation may block. "
           "A lower value improves responsiveness at the expense of "
-          "slightly higher CPU usage.",
-	  1, 60*1000, 100 },
+          "slightly higher CPU usage. **Deprecated**",
+	  1, 60*1000, 1000 },
 	{ _RK_GLOBAL, "socket.send.buffer.bytes", _RK_C_INT,
 	  _RK(socket_sndbuf_size),
 	  "Broker socket send buffer size. System default is used if 0.",
@@ -286,11 +320,19 @@ static const struct rd_kafka_property rd_kafka_properties[] = {
           _RK(log_level),
           "Logging level (syslog(3) levels)",
           0, 7, 6 },
+        { _RK_GLOBAL, "log.queue", _RK_C_BOOL, _RK(log_queue),
+          "Disable spontaneous log_cb from internal librdkafka "
+          "threads, instead enqueue log messages on queue set with "
+          "`rd_kafka_set_log_queue()` and serve log callbacks or "
+          "events through the standard poll APIs. "
+          "**NOTE**: Log messages will linger in a temporary queue "
+          "until the log queue has been set.",
+          0, 1, 0 },
 	{ _RK_GLOBAL, "log.thread.name", _RK_C_BOOL,
 	  _RK(log_thread_name),
 	  "Print internal thread name in log messages "
 	  "(useful for debugging librdkafka internals)",
-	  0, 1, ENABLE_DEVEL },
+	  0, 1, 1 },
 	{ _RK_GLOBAL, "log.connection.close", _RK_C_BOOL,
 	  _RK(log_connection_close),
 	  "Log broker disconnects. "
@@ -344,11 +386,12 @@ static const struct rd_kafka_property rd_kafka_properties[] = {
 	{ _RK_GLOBAL, "api.version.request", _RK_C_BOOL,
 	  _RK(api_version_request),
 	  "Request broker's supported API versions to adjust functionality to "
-	  "available protocol features. If set to false the fallback version "
+	  "available protocol features. If set to false, or the "
+          "ApiVersionRequest fails, the fallback version "
 	  "`broker.version.fallback` will be used. "
 	  "**NOTE**: Depends on broker version >=0.10.0. If the request is not "
 	  "supported by (an older) broker the `broker.version.fallback` fallback is used.",
-	  0, 1, 0 },
+	  0, 1, 1 },
 	{ _RK_GLOBAL, "api.version.fallback.ms", _RK_C_INT,
 	  _RK(api_version_fallback_ms),
 	  "Dictates how long the `broker.version.fallback` fallback is used "
@@ -381,11 +424,9 @@ static const struct rd_kafka_property rd_kafka_properties[] = {
 #if WITH_SSL
 			{ RD_KAFKA_PROTO_SSL, "ssl" },
 #endif
-#if WITH_SASL
 			{ RD_KAFKA_PROTO_SASL_PLAINTEXT, "sasl_plaintext" },
 #if WITH_SSL
 			{ RD_KAFKA_PROTO_SASL_SSL, "sasl_ssl" },
-#endif
 #endif
 			{ 0, NULL }
 		} },
@@ -422,11 +463,10 @@ static const struct rd_kafka_property rd_kafka_properties[] = {
 	},
 #endif /* WITH_SSL */
 
-#if WITH_SASL
 	{_RK_GLOBAL,"sasl.mechanisms", _RK_C_STR,
 	 _RK(sasl.mechanisms),
 	 "SASL mechanism to use for authentication. "
-	 "Supported: GSSAPI, PLAIN. "
+	 "Supported: GSSAPI, PLAIN, SCRAM-SHA-256, SCRAM-SHA-512. "
 	 "**NOTE**: Despite the name only one mechanism must be configured.",
 	 .sdef = "GSSAPI",
 	 .validate = rd_kafka_conf_validate_single },
@@ -438,6 +478,7 @@ static const struct rd_kafka_property rd_kafka_properties[] = {
 	  _RK(sasl.principal),
 	  "This client's Kerberos principal name.",
 	  .sdef = "kafkaclient" },
+#ifndef _MSC_VER
 	{ _RK_GLOBAL, "sasl.kerberos.kinit.cmd", _RK_C_STR,
 	  _RK(sasl.kinit_cmd),
 	  "Full kerberos kinit command string, %{config.prop.name} is replaced "
@@ -455,14 +496,37 @@ static const struct rd_kafka_property rd_kafka_properties[] = {
 	  _RK(sasl.relogin_min_time),
 	  "Minimum time in milliseconds between key refresh attempts.",
 	  1, 86400*1000, 60*1000 },
+#endif
 	{ _RK_GLOBAL, "sasl.username", _RK_C_STR,
 	  _RK(sasl.username),
-	  "SASL username for use with the PLAIN mechanism" },
+	  "SASL username for use with the PLAIN and SASL-SCRAM-.. mechanisms" },
 	{ _RK_GLOBAL, "sasl.password", _RK_C_STR,
 	  _RK(sasl.password),
-	  "SASL password for use with the PLAIN mechanism" },
+	  "SASL password for use with the PLAIN and SASL-SCRAM-.. mechanism" },
+
+#if WITH_PLUGINS
+        /* Plugins */
+        { _RK_GLOBAL, "plugin.library.paths", _RK_C_STR,
+          _RK(plugin_paths),
+          "List of plugin libaries to load (; separated). "
+          "The library search path is platform dependent (see dlopen(3) for Unix and LoadLibrary() for Windows). If no filename extension is specified the "
+          "platform-specific extension (such as .dll or .so) will be appended automatically.",
+          .set = rd_kafka_plugins_conf_set },
 #endif
 
+        /* Interceptors are added through specific API and not exposed
+         * as configuration properties.
+         * The interceptor property must be defined after plugin.library.paths
+         * so that the plugin libraries are properly loaded before
+         * interceptors are configured when duplicating configuration objects.*/
+        { _RK_GLOBAL, "interceptors", _RK_C_INTERNAL,
+          _RK(interceptors),
+          "Interceptors added through rd_kafka_conf_interceptor_add_..() "
+          "and any configuration handled by interceptors.",
+          .ctor = rd_kafka_conf_interceptor_ctor,
+          .dtor = rd_kafka_conf_interceptor_dtor,
+          .copy = rd_kafka_conf_interceptor_copy },
+
         /* Global client group properties */
         { _RK_GLOBAL|_RK_CGRP, "group.id", _RK_C_STR,
           _RK(group_id_str),
@@ -501,7 +565,8 @@ static const struct rd_kafka_property rd_kafka_properties[] = {
         { _RK_GLOBAL|_RK_CONSUMER, "auto.commit.interval.ms", _RK_C_INT,
 	  _RK(auto_commit_interval_ms),
 	  "The frequency in milliseconds that the consumer offsets "
-	  "are committed (written) to offset storage. (0 = disable)",
+	  "are committed (written) to offset storage. (0 = disable). "
+          "This setting is used by the high-level consumer.",
           0, 86400*1000, 5*1000 },
         { _RK_GLOBAL|_RK_CONSUMER, "enable.auto.offset.store", _RK_C_BOOL,
           _RK(enable_auto_offset_store),
@@ -574,7 +639,12 @@ static const struct rd_kafka_property rd_kafka_properties[] = {
 	  "Emit RD_KAFKA_RESP_ERR__PARTITION_EOF event whenever the "
 	  "consumer reaches the end of a partition.",
 	  0, 1, 1 },
-
+        { _RK_GLOBAL|_RK_CONSUMER, "check.crcs", _RK_C_BOOL,
+          _RK(check_crcs),
+          "Verify CRC32 of consumed messages, ensuring no on-the-wire or "
+          "on-disk corruption to the messages occurred. This check comes "
+          "at slightly increased CPU usage.",
+          0, 1, 0 },
 	/* Global producer properties */
 	{ _RK_GLOBAL|_RK_PRODUCER, "queue.buffering.max.messages", _RK_C_INT,
 	  _RK(queue_buffering_max_msgs),
@@ -588,7 +658,9 @@ static const struct rd_kafka_property rd_kafka_properties[] = {
 	  _RK(buffering_max_ms),
 	  "Maximum time, in milliseconds, for buffering data "
 	  "on the producer queue.",
-	  1, 900*1000, 1000 },
+	  0, 900*1000, 0 },
+        { _RK_GLOBAL|_RK_PRODUCER, "linger.ms", _RK_C_ALIAS,
+          .sdef = "queue.buffering.max.ms" },
 	{ _RK_GLOBAL|_RK_PRODUCER, "message.send.max.retries", _RK_C_INT,
 	  _RK(max_retries),
 	  "How many times to retry sending a failing MessageSet. "
@@ -614,9 +686,7 @@ static const struct rd_kafka_property rd_kafka_properties[] = {
 #if WITH_SNAPPY
 			{ RD_KAFKA_COMPRESSION_SNAPPY, "snappy" },
 #endif
-#if WITH_LZ4
                         { RD_KAFKA_COMPRESSION_LZ4, "lz4" },
-#endif
 			{ 0 }
 		} },
 	{ _RK_GLOBAL|_RK_PRODUCER, "batch.num.messages", _RK_C_INT,
@@ -696,10 +766,7 @@ static const struct rd_kafka_property rd_kafka_properties[] = {
 #if WITH_SNAPPY
 		  { RD_KAFKA_COMPRESSION_SNAPPY, "snappy" },
 #endif
-#if WITH_LZ4
 		  { RD_KAFKA_COMPRESSION_LZ4, "lz4" },
-#endif
-
 		  { RD_KAFKA_COMPRESSION_INHERIT, "inherit" },
 		  { 0 }
 		} },
@@ -725,7 +792,8 @@ static const struct rd_kafka_property rd_kafka_properties[] = {
 	{ _RK_TOPIC|_RK_CONSUMER, "auto.commit.interval.ms", _RK_C_INT,
 	  _RKT(auto_commit_interval_ms),
 	  "The frequency in milliseconds that the consumer offsets "
-	  "are committed (written) to offset storage.",
+	  "are committed (written) to offset storage. "
+          "This setting is used by the low-level legacy consumer.",
 	  10, 86400*1000, 60*1000 },
 	{ _RK_TOPIC|_RK_CONSUMER, "auto.offset.reset", _RK_C_S2I,
 	  _RKT(auto_offset_reset),
@@ -766,8 +834,8 @@ static const struct rd_kafka_property rd_kafka_properties[] = {
           "'file' - local file store (offset.store.path, et.al), "
           "'broker' - broker commit store "
           "(requires \"group.id\" to be configured and "
-	  "Apache Kafka 0.8.2 or later on the broker.).",
-          .vdef = RD_KAFKA_OFFSET_METHOD_BROKER, /* FIXME: warn about default change */
+          "Apache Kafka 0.8.2 or later on the broker.).",
+          .vdef = RD_KAFKA_OFFSET_METHOD_BROKER,
           .s2i = {
                         { RD_KAFKA_OFFSET_METHOD_FILE, "file" },
                         { RD_KAFKA_OFFSET_METHOD_BROKER, "broker" }
@@ -784,19 +852,44 @@ static const struct rd_kafka_property rd_kafka_properties[] = {
 };
 
 
-typedef enum {
-	_PROP_SET_REPLACE,  /* Replace current value (default) */
-	_PROP_SET_ADD,      /* Add value (S2F) */
-	_PROP_SET_DEL      /* Remove value (S2F) */
-} prop_set_mode_t;
-
 static rd_kafka_conf_res_t
 rd_kafka_anyconf_set_prop0 (int scope, void *conf,
 			    const struct rd_kafka_property *prop,
-			    const char *istr, int ival, prop_set_mode_t set_mode,
+			    const char *istr, int ival, rd_kafka_conf_set_mode_t set_mode,
                             char *errstr, size_t errstr_size) {
+        rd_kafka_conf_res_t res;
+
+#define _RK_PTR(TYPE,BASE,OFFSET)  (TYPE)(void *)(((char *)(BASE))+(OFFSET))
+
+        /* Try interceptors first (only for GLOBAL config) */
+        if (scope & _RK_GLOBAL) {
+                if (prop->type == _RK_C_PTR || prop->type == _RK_C_INTERNAL)
+                        res = RD_KAFKA_CONF_UNKNOWN;
+                else
+                        res = rd_kafka_interceptors_on_conf_set(conf,
+                                                                prop->name,
+                                                                istr,
+                                                                errstr,
+                                                                errstr_size);
+                if (res != RD_KAFKA_CONF_UNKNOWN)
+                        return res;
+        }
+
+
+        if (prop->set) {
+                /* Custom setter */
+                rd_kafka_conf_res_t res;
+
+                res = prop->set(scope, conf, prop->name, istr,
+                                _RK_PTR(void *, conf, prop->offset),
+                                set_mode, errstr, errstr_size);
+
+                if (res != RD_KAFKA_CONF_OK)
+                        return res;
+
+                /* FALLTHRU so that property value is set. */
+        }
 
-#define _RK_PTR(TYPE,BASE,OFFSET)  (TYPE)(((char *)(BASE))+(OFFSET))
 	switch (prop->type)
 	{
 	case _RK_C_STR:
@@ -836,13 +929,13 @@ rd_kafka_anyconf_set_prop0 (int scope, void *conf,
 		if (prop->type == _RK_C_S2F) {
 			switch (set_mode)
 			{
-			case _PROP_SET_REPLACE:
+			case _RK_CONF_PROP_SET_REPLACE:
 				*val = ival;
 				break;
-			case _PROP_SET_ADD:
+			case _RK_CONF_PROP_SET_ADD:
 				*val |= ival;
 				break;
-			case _PROP_SET_DEL:
+			case _RK_CONF_PROP_SET_DEL:
 				*val &= ~ival;
 				break;
 			}
@@ -868,7 +961,7 @@ rd_kafka_anyconf_set_prop0 (int scope, void *conf,
 			if (!(*plist =
 			      rd_kafka_pattern_list_new(istr,
 							errstr,
-                                                        (int)errstr_size)))
+							(int)errstr_size)))
 				return RD_KAFKA_CONF_INVALID;
 		} else
 			*plist = NULL;
@@ -876,6 +969,10 @@ rd_kafka_anyconf_set_prop0 (int scope, void *conf,
                 return RD_KAFKA_CONF_OK;
         }
 
+        case _RK_C_INTERNAL:
+                /* Probably handled by setter */
+                return RD_KAFKA_CONF_OK;
+
 	default:
 		rd_kafka_assert(NULL, !*"unknown conf type");
 	}
@@ -942,7 +1039,7 @@ rd_kafka_anyconf_set_prop (int scope, void *conf,
 		}
 
 		return rd_kafka_anyconf_set_prop0(scope, conf, prop, value, 0,
-						  _PROP_SET_REPLACE,
+						  _RK_CONF_PROP_SET_REPLACE,
                                                   errstr, errstr_size);
 
 	case _RK_C_PTR:
@@ -975,8 +1072,8 @@ rd_kafka_anyconf_set_prop (int scope, void *conf,
 			return RD_KAFKA_CONF_INVALID;
 		}
 
-		rd_kafka_anyconf_set_prop0(scope, conf, prop, NULL, ival,
-					   _PROP_SET_REPLACE,
+		rd_kafka_anyconf_set_prop0(scope, conf, prop, value, ival,
+					   _RK_CONF_PROP_SET_REPLACE,
                                            errstr, errstr_size);
 		return RD_KAFKA_CONF_OK;
 
@@ -1019,8 +1116,8 @@ rd_kafka_anyconf_set_prop (int scope, void *conf,
 			return RD_KAFKA_CONF_INVALID;
 		}
 
-		rd_kafka_anyconf_set_prop0(scope, conf, prop, NULL, ival,
-					   _PROP_SET_REPLACE,
+		rd_kafka_anyconf_set_prop0(scope, conf, prop, value, ival,
+					   _RK_CONF_PROP_SET_REPLACE,
                                            errstr, errstr_size);
 		return RD_KAFKA_CONF_OK;
 	}
@@ -1042,7 +1139,7 @@ rd_kafka_anyconf_set_prop (int scope, void *conf,
 		next = value;
 		while (next && *next) {
 			const char *s, *t;
-			prop_set_mode_t set_mode = _PROP_SET_ADD; /* S2F */
+			rd_kafka_conf_set_mode_t set_mode = _RK_CONF_PROP_SET_ADD; /* S2F */
 
 			s = next;
 
@@ -1068,10 +1165,10 @@ rd_kafka_anyconf_set_prop (int scope, void *conf,
 			/* S2F: +/- prefix */
 			if (prop->type == _RK_C_S2F) {
 				if (*s == '+') {
-					set_mode = _PROP_SET_ADD;
+					set_mode = _RK_CONF_PROP_SET_ADD;
 					s++;
 				} else if (*s == '-') {
-					set_mode = _PROP_SET_DEL;
+					set_mode = _RK_CONF_PROP_SET_DEL;
 					s++;
 				}
 			}
@@ -1095,7 +1192,8 @@ rd_kafka_anyconf_set_prop (int scope, void *conf,
 					continue;
 
 				rd_kafka_anyconf_set_prop0(scope, conf, prop,
-							   NULL, new_val, set_mode,
+                                                           value, new_val,
+                                                           set_mode,
                                                            errstr, errstr_size);
 
 				if (prop->type == _RK_C_S2F) {
@@ -1141,11 +1239,15 @@ static void rd_kafka_defaultconf_set (int scope, void *conf) {
 		if (prop->type == _RK_C_ALIAS)
 			continue;
 
+                if (prop->ctor)
+                        prop->ctor(scope, conf);
+
 		if (prop->sdef || prop->vdef || prop->pdef)
 			rd_kafka_anyconf_set_prop0(scope, conf, prop,
 						   prop->sdef ?
                                                    prop->sdef : prop->pdef,
-                                                   prop->vdef, _PROP_SET_REPLACE,
+                                                   prop->vdef,
+                                                   _RK_CONF_PROP_SET_REPLACE,
                                                    NULL, 0);
 	}
 }
@@ -1169,6 +1271,7 @@ static int rd_kafka_anyconf_set (int scope, void *conf,
 				 char *errstr, size_t errstr_size) {
 	char estmp[1];
 	const struct rd_kafka_property *prop;
+        rd_kafka_conf_res_t res;
 
 	if (!errstr) {
 		errstr = estmp;
@@ -1178,6 +1281,19 @@ static int rd_kafka_anyconf_set (int scope, void *conf,
 	if (value && !*value)
 		value = NULL;
 
+        /* Try interceptors first (only for GLOBAL config for now) */
+        if (scope & _RK_GLOBAL) {
+                res = rd_kafka_interceptors_on_conf_set(
+                        (rd_kafka_conf_t *)conf, name, value,
+                        errstr, errstr_size);
+                /* Handled (successfully or not) by interceptor. */
+                if (res != RD_KAFKA_CONF_UNKNOWN)
+                        return res;
+        }
+
+        /* Then global config */
+
+
 	for (prop = rd_kafka_properties ; prop->name ; prop++) {
 
 		if (!(prop->scope & scope))
@@ -1203,11 +1319,27 @@ static int rd_kafka_anyconf_set (int scope, void *conf,
 
 
 rd_kafka_conf_res_t rd_kafka_conf_set (rd_kafka_conf_t *conf,
-				       const char *name,
-				       const char *value,
-				       char *errstr, size_t errstr_size) {
-	return rd_kafka_anyconf_set(_RK_GLOBAL, conf, name, value,
-				    errstr, errstr_size);
+                                       const char *name,
+                                       const char *value,
+                                       char *errstr, size_t errstr_size) {
+        rd_kafka_conf_res_t res;
+
+        res = rd_kafka_anyconf_set(_RK_GLOBAL, conf, name, value,
+                                   errstr, errstr_size);
+        if (res != RD_KAFKA_CONF_UNKNOWN)
+                return res;
+
+        /* Fallthru:
+         * If the global property was unknown, try setting it on the
+         * default topic config. */
+        if (!conf->topic_conf) {
+                /* Create topic config, might be over-written by application
+                 * later. */
+                conf->topic_conf = rd_kafka_topic_conf_new();
+        }
+
+        return rd_kafka_topic_conf_set(conf->topic_conf, name, value,
+                                       errstr, errstr_size);
 }
 
 
@@ -1223,15 +1355,21 @@ rd_kafka_conf_res_t rd_kafka_topic_conf_set (rd_kafka_topic_conf_t *conf,
 }
 
 
-static void rd_kafka_anyconf_clear (void *conf,
+static void rd_kafka_anyconf_clear (int scope, void *conf,
 				    const struct rd_kafka_property *prop) {
 	switch (prop->type)
 	{
 	case _RK_C_STR:
 	{
 		char **str = _RK_PTR(char **, conf, prop->offset);
+
 		if (*str) {
-			rd_free(*str);
+                        if (prop->set) {
+                                prop->set(scope, conf, prop->name, NULL, *str,
+                                          _RK_CONF_PROP_SET_DEL, NULL, 0);
+                                /* FALLTHRU */
+                        }
+                        rd_free(*str);
 			*str = NULL;
 		}
 	}
@@ -1278,16 +1416,23 @@ static void rd_kafka_anyconf_clear (void *conf,
 		break;
 	}
 
+        if (prop->dtor)
+                prop->dtor(scope, conf);
+
 }
 
 void rd_kafka_anyconf_destroy (int scope, void *conf) {
 	const struct rd_kafka_property *prop;
 
+        /* Call on_conf_destroy() interceptors */
+        if (scope == _RK_GLOBAL)
+                rd_kafka_interceptors_on_conf_destroy(conf);
+
 	for (prop = rd_kafka_properties; prop->name ; prop++) {
 		if (!(prop->scope & scope))
 			continue;
 
-		rd_kafka_anyconf_clear(conf, prop);
+		rd_kafka_anyconf_clear(scope, conf, prop);
 	}
 }
 
@@ -1305,12 +1450,17 @@ void rd_kafka_topic_conf_destroy (rd_kafka_topic_conf_t *topic_conf) {
 
 
 
-static void rd_kafka_anyconf_copy (int scope, void *dst, const void *src) {
+static void rd_kafka_anyconf_copy (int scope, void *dst, const void *src,
+                                   size_t filter_cnt, const char **filter) {
 	const struct rd_kafka_property *prop;
 
 	for (prop = rd_kafka_properties ; prop->name ; prop++) {
 		const char *val = NULL;
 		int ival = 0;
+                char *valstr;
+                size_t valsz;
+                size_t fi;
+                size_t nlen;
 
 		if (!(prop->scope & scope))
 			continue;
@@ -1318,6 +1468,17 @@ static void rd_kafka_anyconf_copy (int scope, void *dst, const void *src) {
 		if (prop->type == _RK_C_ALIAS)
 			continue;
 
+                /* Apply filter, if any. */
+                nlen = strlen(prop->name);
+                for (fi = 0 ; fi < filter_cnt ; fi++) {
+                        size_t flen = strlen(filter[fi]);
+                        if (nlen >= flen &&
+                            !strncmp(filter[fi], prop->name, flen))
+                                break;
+                }
+                if (fi < filter_cnt)
+                        continue; /* Filter matched */
+
 		switch (prop->type)
 		{
 		case _RK_C_STR:
@@ -1326,7 +1487,8 @@ static void rd_kafka_anyconf_copy (int scope, void *dst, const void *src) {
 
                         if (!strcmp(prop->name, "default_topic_conf") && val)
                                 val = (void *)rd_kafka_topic_conf_dup(
-                                        (const rd_kafka_topic_conf_t *)val);
+                                        (const rd_kafka_topic_conf_t *)
+                                        (void *)val);
 			break;
                 case _RK_C_KSTR:
                 {
@@ -1342,6 +1504,13 @@ static void rd_kafka_anyconf_copy (int scope, void *dst, const void *src) {
 		case _RK_C_S2I:
 		case _RK_C_S2F:
 			ival = *_RK_PTR(const int *, src, prop->offset);
+
+                        /* Get string representation of configuration value. */
+                        valsz = 0;
+                        rd_kafka_anyconf_get0(src, prop, NULL, &valsz);
+                        valstr = rd_alloca(valsz);
+                        rd_kafka_anyconf_get0(src, prop, valstr, &valsz);
+                        val = valstr;
 			break;
                 case _RK_C_PATLIST:
                 {
@@ -1352,12 +1521,21 @@ static void rd_kafka_anyconf_copy (int scope, void *dst, const void *src) {
 				val = (*plist)->rkpl_orig;
                         break;
                 }
+                case _RK_C_INTERNAL:
+                        /* Handled by ->copy() below. */
+                        break;
 		default:
 			continue;
 		}
 
-		rd_kafka_anyconf_set_prop0(scope, dst, prop, val, ival,
-                                           _PROP_SET_REPLACE, NULL, 0);
+                if (prop->copy)
+                        prop->copy(scope, dst, src,
+                                   _RK_PTR(void *, dst, prop->offset),
+                                   _RK_PTR(const void *, src, prop->offset),
+                                   filter_cnt, filter);
+
+                rd_kafka_anyconf_set_prop0(scope, dst, prop, val, ival,
+                                           _RK_CONF_PROP_SET_REPLACE, NULL, 0);
 	}
 }
 
@@ -1365,7 +1543,21 @@ static void rd_kafka_anyconf_copy (int scope, void *dst, const void *src) {
 rd_kafka_conf_t *rd_kafka_conf_dup (const rd_kafka_conf_t *conf) {
 	rd_kafka_conf_t *new = rd_kafka_conf_new();
 
-	rd_kafka_anyconf_copy(_RK_GLOBAL, new, conf);
+        rd_kafka_interceptors_on_conf_dup(new, conf, 0, NULL);
+
+        rd_kafka_anyconf_copy(_RK_GLOBAL, new, conf, 0, NULL);
+
+	return new;
+}
+
+rd_kafka_conf_t *rd_kafka_conf_dup_filter (const rd_kafka_conf_t *conf,
+                                           size_t filter_cnt,
+                                           const char **filter) {
+	rd_kafka_conf_t *new = rd_kafka_conf_new();
+
+        rd_kafka_interceptors_on_conf_dup(new, conf, filter_cnt, filter);
+
+        rd_kafka_anyconf_copy(_RK_GLOBAL, new, conf, filter_cnt, filter);
 
 	return new;
 }
@@ -1375,7 +1567,7 @@ rd_kafka_topic_conf_t *rd_kafka_topic_conf_dup (const rd_kafka_topic_conf_t
 						*conf) {
 	rd_kafka_topic_conf_t *new = rd_kafka_topic_conf_new();
 
-	rd_kafka_anyconf_copy(_RK_TOPIC, new, conf);
+	rd_kafka_anyconf_copy(_RK_TOPIC, new, conf, 0, NULL);
 
 	return new;
 }
@@ -1723,7 +1915,15 @@ rd_kafka_conf_res_t rd_kafka_topic_conf_get (const rd_kafka_topic_conf_t *conf,
 rd_kafka_conf_res_t rd_kafka_conf_get (const rd_kafka_conf_t *conf,
                                        const char *name,
                                        char *dest, size_t *dest_size) {
-        return rd_kafka_anyconf_get(_RK_GLOBAL, conf, name, dest, dest_size);
+        rd_kafka_conf_res_t res;
+        res = rd_kafka_anyconf_get(_RK_GLOBAL, conf, name, dest, dest_size);
+        if (res != RD_KAFKA_CONF_UNKNOWN || !conf->topic_conf)
+                return res;
+
+        /* Fallthru:
+         * If the global property was unknown, try getting it from the
+         * default topic config, if any. */
+        return rd_kafka_topic_conf_get(conf->topic_conf, name, dest, dest_size);
 }
 
 
