@@ -31,6 +31,7 @@
 #include "rdkafka_msg.h"
 #include "rdkafka_topic.h"
 #include "rdkafka_partition.h"
+#include "rdkafka_interceptor.h"
 #include "rdcrc32.h"
 #include "rdrand.h"
 #include "rdtime.h"
@@ -68,6 +69,7 @@ void rd_kafka_msg_destroy (rd_kafka_t *rk, rd_kafka_msg_t *rkm) {
  *
  * @returns the new message
  */
+static
 rd_kafka_msg_t *rd_kafka_msg_new00 (rd_kafka_itopic_t *rkt,
 				    int32_t partition,
 				    int msgflags,
@@ -144,8 +146,8 @@ static rd_kafka_msg_t *rd_kafka_msg_new0 (rd_kafka_itopic_t *rkt,
                                           const void *key, size_t keylen,
                                           void *msg_opaque,
                                           rd_kafka_resp_err_t *errp,
-					  int *errnop,
-					  rd_ts_t utc_now,
+                                          int *errnop,
+                                          int64_t timestamp,
                                           rd_ts_t now) {
 	rd_kafka_msg_t *rkm;
 
@@ -176,8 +178,13 @@ static rd_kafka_msg_t *rd_kafka_msg_new0 (rd_kafka_itopic_t *rkt,
 				 msgflags|RD_KAFKA_MSG_F_ACCOUNT /* curr_msgs_add() */,
 				 payload, len, key, keylen, msg_opaque);
 
-	rkm->rkm_timestamp  = utc_now / 1000;
-	rkm->rkm_tstype     = RD_KAFKA_TIMESTAMP_CREATE_TIME;
+        if (timestamp)
+                rkm->rkm_timestamp  = timestamp;
+        else
+                rkm->rkm_timestamp = rd_uclock()/1000;
+        rkm->rkm_tstype     = RD_KAFKA_TIMESTAMP_CREATE_TIME;
+
+        rkm->rkm_ts_enq = now;
 
 	if (rkt->rkt_conf.message_timeout_ms == 0) {
 		rkm->rkm_ts_timeout = INT64_MAX;
@@ -186,19 +193,24 @@ static rd_kafka_msg_t *rd_kafka_msg_new0 (rd_kafka_itopic_t *rkt,
 			rkt->rkt_conf.message_timeout_ms * 1000;
 	}
 
+        /* Call interceptor chain for on_send */
+        rd_kafka_interceptors_on_send(rkt->rkt_rk, &rkm->rkm_rkmessage);
+
         return rkm;
 }
 
 
 /**
- * Produce: creates a new message, runs the partitioner and enqueues
- *          into on the selected partition.
+ * @brief Produce: creates a new message, runs the partitioner and enqueues
+ *        into on the selected partition.
  *
- * Returns 0 on success or -1 on error.
+ * @returns 0 on success or -1 on error.
  *
  * If the function returns -1 and RD_KAFKA_MSG_F_FREE was specified, then
  * the memory associated with the payload is still the caller's
  * responsibility.
+ *
+ * @locks none
  */
 int rd_kafka_msg_new (rd_kafka_itopic_t *rkt, int32_t force_partition,
 		      int msgflags,
@@ -212,7 +224,7 @@ int rd_kafka_msg_new (rd_kafka_itopic_t *rkt, int32_t force_partition,
         /* Create message */
         rkm = rd_kafka_msg_new0(rkt, force_partition, msgflags, 
                                 payload, len, key, keylen, msg_opaque,
-				&err, &errnox, rd_uclock(), rd_clock());
+                                &err, &errnox, 0, rd_clock());
         if (unlikely(!rkm)) {
                 /* errno is already set by msg_new() */
 		rd_kafka_set_last_error(err, errnox);
@@ -227,6 +239,11 @@ int rd_kafka_msg_new (rd_kafka_itopic_t *rkt, int32_t force_partition,
 		return 0;
 	}
 
+        /* Interceptor: unroll failing messages by triggering on_ack.. */
+        rkm->rkm_err = err;
+        rd_kafka_interceptors_on_acknowledgement(rkt->rkt_rk,
+                                                 &rkm->rkm_rkmessage);
+
 	/* Handle partitioner failures: it only fails when the application
 	 * attempts to force a destination partition that does not exist
 	 * in the cluster.  Note we must clear the RD_KAFKA_MSG_F_FREE
@@ -248,7 +265,112 @@ int rd_kafka_msg_new (rd_kafka_itopic_t *rkt, int32_t force_partition,
 }
 
 
+rd_kafka_resp_err_t rd_kafka_producev (rd_kafka_t *rk, ...) {
+        va_list ap;
+        rd_kafka_msg_t s_rkm = {
+                /* Message defaults */
+                .rkm_partition = RD_KAFKA_PARTITION_UA,
+                .rkm_timestamp = 0, /* current time */
+        };
+        rd_kafka_msg_t *rkm = &s_rkm;
+        rd_kafka_vtype_t vtype;
+        rd_kafka_topic_t *app_rkt;
+        shptr_rd_kafka_itopic_t *s_rkt = NULL;
+        rd_kafka_itopic_t *rkt;
+        rd_kafka_resp_err_t err = RD_KAFKA_RESP_ERR_NO_ERROR;
 
+        va_start(ap, rk);
+        while ((vtype = va_arg(ap, rd_kafka_vtype_t)) != RD_KAFKA_VTYPE_END) {
+                switch (vtype)
+                {
+                case RD_KAFKA_VTYPE_TOPIC:
+                        s_rkt = rd_kafka_topic_new0(rk,
+                                                    va_arg(ap, const char *),
+                                                    NULL, NULL, 1);
+                        break;
+
+                case RD_KAFKA_VTYPE_RKT:
+                        app_rkt = va_arg(ap, rd_kafka_topic_t *);
+                        s_rkt = rd_kafka_topic_keep(
+                                rd_kafka_topic_a2i(app_rkt));
+                        break;
+
+                case RD_KAFKA_VTYPE_PARTITION:
+                        rkm->rkm_partition = va_arg(ap, int32_t);
+                        break;
+
+                case RD_KAFKA_VTYPE_VALUE:
+                        rkm->rkm_payload = va_arg(ap, void *);
+                        rkm->rkm_len = va_arg(ap, size_t);
+                        break;
+
+                case RD_KAFKA_VTYPE_KEY:
+                        rkm->rkm_key = va_arg(ap, void *);
+                        rkm->rkm_key_len = va_arg(ap, size_t);
+                        break;
+
+                case RD_KAFKA_VTYPE_OPAQUE:
+                        rkm->rkm_opaque = va_arg(ap, void *);
+                        break;
+
+                case RD_KAFKA_VTYPE_MSGFLAGS:
+                        rkm->rkm_flags = va_arg(ap, int);
+                        break;
+
+                case RD_KAFKA_VTYPE_TIMESTAMP:
+                        rkm->rkm_timestamp = va_arg(ap, int64_t);
+                        break;
+
+                default:
+                        err = RD_KAFKA_RESP_ERR__INVALID_ARG;
+                        break;
+                }
+        }
+
+        va_end(ap);
+
+        if (unlikely(!s_rkt))
+                return RD_KAFKA_RESP_ERR__INVALID_ARG;
+
+        rkt = rd_kafka_topic_s2i(s_rkt);
+
+        if (likely(!err))
+                rkm = rd_kafka_msg_new0(rkt,
+                                        rkm->rkm_partition,
+                                        rkm->rkm_flags,
+                                        rkm->rkm_payload, rkm->rkm_len,
+                                        rkm->rkm_key, rkm->rkm_key_len,
+                                        rkm->rkm_opaque,
+                                        &err, NULL,
+                                        rkm->rkm_timestamp, rd_clock());
+
+        if (unlikely(err))
+                return err;
+
+        /* Partition the message */
+        err = rd_kafka_msg_partitioner(rkt, rkm, 1);
+        if (unlikely(err)) {
+                /* Handle partitioner failures: it only fails when
+                 * the application attempts to force a destination
+                 * partition that does not exist in the cluster. */
+
+                /* Interceptors: Unroll on_send by on_ack.. */
+                rkm->rkm_err = err;
+                rd_kafka_interceptors_on_acknowledgement(rk,
+                                                         &rkm->rkm_rkmessage);
+
+                /* Note we must clear the RD_KAFKA_MSG_F_FREE
+                 * flag since our contract says we don't free the payload on
+                 * failure. */
+                rkm->rkm_flags &= ~RD_KAFKA_MSG_F_FREE;
+
+                rd_kafka_msg_destroy(rk, rkm);
+        }
+
+        rd_kafka_topic_destroy0(s_rkt);
+
+        return err;
+}
 
 /**
  * Produce a batch of messages.
@@ -260,7 +382,7 @@ int rd_kafka_produce_batch (rd_kafka_topic_t *app_rkt, int32_t partition,
                             rd_kafka_message_t *rkmessages, int message_cnt) {
         rd_kafka_msgq_t tmpq = RD_KAFKA_MSGQ_INITIALIZER(tmpq);
         int i;
-	rd_ts_t utc_now = rd_uclock();
+	int64_t utc_now = rd_uclock() / 1000;
         rd_ts_t now = rd_clock();
         int good = 0;
         rd_kafka_resp_err_t all_err = 0;
@@ -306,6 +428,10 @@ int rd_kafka_produce_batch (rd_kafka_topic_t *app_rkt, int32_t partition,
                                                          0/*already locked*/);
 
                         if (unlikely(rkmessages[i].err)) {
+                                /* Interceptors: Unroll on_send by on_ack.. */
+                                rd_kafka_interceptors_on_acknowledgement(
+                                        rkt->rkt_rk, &rkmessages[i]);
+
                                 rd_kafka_msg_destroy(rkt->rkt_rk, rkm);
                                 continue;
                         }
@@ -522,3 +648,153 @@ int rd_kafka_msg_partitioner (rd_kafka_itopic_t *rkt, rd_kafka_msg_t *rkm,
 	rd_kafka_toppar_destroy(s_rktp_new); /* from _get() */
 	return 0;
 }
+
+
+
+
+/**
+ * @name Public message type (rd_kafka_message_t)
+ */
+void rd_kafka_message_destroy (rd_kafka_message_t *rkmessage) {
+        rd_kafka_op_t *rko;
+
+        if (likely((rko = (rd_kafka_op_t *)rkmessage->_private) != NULL))
+                rd_kafka_op_destroy(rko);
+        else {
+                rd_kafka_msg_t *rkm = rd_kafka_message2msg(rkmessage);
+                rd_kafka_msg_destroy(NULL, rkm);
+        }
+}
+
+
+rd_kafka_message_t *rd_kafka_message_new (void) {
+        rd_kafka_msg_t *rkm = rd_calloc(1, sizeof(*rkm));
+        return (rd_kafka_message_t *)rkm;
+}
+
+
+/**
+ * @brief Set up a rkmessage from an rko for passing to the application.
+ * @remark Will trigger on_consume() interceptors if any.
+ */
+static rd_kafka_message_t *
+rd_kafka_message_setup (rd_kafka_op_t *rko, rd_kafka_message_t *rkmessage) {
+        rd_kafka_itopic_t *rkt;
+        rd_kafka_toppar_t *rktp = NULL;
+
+        if (rko->rko_type == RD_KAFKA_OP_DR) {
+                rkt = rd_kafka_topic_s2i(rko->rko_u.dr.s_rkt);
+        } else {
+                if (rko->rko_rktp) {
+                        rktp = rd_kafka_toppar_s2i(rko->rko_rktp);
+                        rkt = rktp->rktp_rkt;
+                } else
+                        rkt = NULL;
+
+                rkmessage->_private = rko;
+        }
+
+
+        if (!rkmessage->rkt && rkt)
+                rkmessage->rkt = rd_kafka_topic_keep_a(rkt);
+
+        if (rktp)
+                rkmessage->partition = rktp->rktp_partition;
+
+        if (!rkmessage->err)
+                rkmessage->err = rko->rko_err;
+
+        /* Call on_consume interceptors */
+        switch (rko->rko_type)
+        {
+        case RD_KAFKA_OP_FETCH:
+                if (!rkmessage->err && rkt)
+                        rd_kafka_interceptors_on_consume(rkt->rkt_rk,
+                                                         rkmessage);
+                break;
+
+        default:
+                break;
+        }
+
+        return rkmessage;
+}
+
+
+
+/**
+ * @brief Get rkmessage from rkm (for EVENT_DR)
+ * @remark Must only be called just prior to passing a dr to the application.
+ */
+rd_kafka_message_t *rd_kafka_message_get_from_rkm (rd_kafka_op_t *rko,
+                                                   rd_kafka_msg_t *rkm) {
+        return rd_kafka_message_setup(rko, &rkm->rkm_rkmessage);
+}
+
+/**
+ * @brief Convert rko to rkmessage
+ * @remark Must only be called just prior to passing a consumed message
+ *         or event to the application.
+ * @remark Will trigger on_consume() interceptors, if any.
+ * @returns a rkmessage (bound to the rko).
+ */
+rd_kafka_message_t *rd_kafka_message_get (rd_kafka_op_t *rko) {
+        rd_kafka_message_t *rkmessage;
+
+        if (!rko)
+                return rd_kafka_message_new(); /* empty */
+
+        switch (rko->rko_type)
+        {
+        case RD_KAFKA_OP_FETCH:
+                /* Use embedded rkmessage */
+                rkmessage = &rko->rko_u.fetch.rkm.rkm_rkmessage;
+                break;
+
+        case RD_KAFKA_OP_ERR:
+        case RD_KAFKA_OP_CONSUMER_ERR:
+                rkmessage = &rko->rko_u.err.rkm.rkm_rkmessage;
+                rkmessage->payload = rko->rko_u.err.errstr;
+                rkmessage->offset  = rko->rko_u.err.offset;
+                break;
+
+        default:
+                rd_kafka_assert(NULL, !*"unhandled optype");
+                RD_NOTREACHED();
+                return NULL;
+        }
+
+        return rd_kafka_message_setup(rko, rkmessage);
+}
+
+
+int64_t rd_kafka_message_timestamp (const rd_kafka_message_t *rkmessage,
+                                    rd_kafka_timestamp_type_t *tstype) {
+        rd_kafka_msg_t *rkm;
+
+        if (rkmessage->err) {
+                if (tstype)
+                        *tstype = RD_KAFKA_TIMESTAMP_NOT_AVAILABLE;
+                return -1;
+        }
+
+        rkm = rd_kafka_message2msg((rd_kafka_message_t *)rkmessage);
+
+        if (tstype)
+                *tstype = rkm->rkm_tstype;
+
+        return rkm->rkm_timestamp;
+}
+
+
+int64_t rd_kafka_message_latency (const rd_kafka_message_t *rkmessage) {
+        rd_kafka_msg_t *rkm;
+
+        rkm = rd_kafka_message2msg((rd_kafka_message_t *)rkmessage);
+
+        if (unlikely(!rkm->rkm_ts_enq))
+                return -1;
+
+        return rd_clock() - rkm->rkm_ts_enq;
+}
+
