@@ -66,7 +66,7 @@ typedef struct rd_kafka_replyq_s {
 
 
 typedef enum {
-        RD_KAFKA_OP_NONE,
+        RD_KAFKA_OP_NONE,     /* No specific type, use OP_CB */
 	RD_KAFKA_OP_FETCH,    /* Kafka thread -> Application */
 	RD_KAFKA_OP_ERR,      /* Kafka thread -> Application */
         RD_KAFKA_OP_CONSUMER_ERR, /* Kafka thread -> Application */
@@ -74,7 +74,6 @@ typedef enum {
 			       * Produce message delivery report */
 	RD_KAFKA_OP_STATS,    /* Kafka thread -> Application */
 
-	RD_KAFKA_OP_METADATA_REQ,  /* any -> Broker thread: request metadata */
         RD_KAFKA_OP_OFFSET_COMMIT, /* any -> toppar's Broker thread */
         RD_KAFKA_OP_NODE_UPDATE,   /* any -> Broker thread: node update */
 
@@ -105,6 +104,9 @@ typedef enum {
 	RD_KAFKA_OP_THROTTLE,        /* Throttle info */
 	RD_KAFKA_OP_NAME,            /* Request name */
 	RD_KAFKA_OP_OFFSET_RESET,    /* Offset reset */
+        RD_KAFKA_OP_METADATA,        /* Metadata response */
+        RD_KAFKA_OP_LOG,             /* Log */
+        RD_KAFKA_OP_WAKEUP,          /* Wake-up signaling */
         RD_KAFKA_OP__END
 } rd_kafka_op_type_t;
 
@@ -114,6 +116,68 @@ typedef enum {
 #define RD_KAFKA_OP_FLAGMASK  (RD_KAFKA_OP_CB | RD_KAFKA_OP_REPLY)
 
 
+/**
+ * @brief Op/queue priority levels.
+ * @remark Since priority levels alter the FIFO order, pay extra attention
+ *         to preserve ordering as deemed necessary.
+ * @remark Priority should only be set on ops destined for application
+ *         facing queues (rk_rep, rkcg_q, etc).
+ */
+typedef enum {
+        RD_KAFKA_PRIO_NORMAL = 0,   /* Normal bulk, messages, DRs, etc. */
+        RD_KAFKA_PRIO_MEDIUM,       /* Prioritize in front of bulk,
+                                     * still at some scale. e.g. logs, .. */
+        RD_KAFKA_PRIO_HIGH,         /* Small scale high priority */
+        RD_KAFKA_PRIO_FLASH         /* Micro scale, immediate delivery. */
+} rd_kafka_op_prio_t;
+
+
+/**
+ * @brief Op handler result
+ *
+ * @remark When returning YIELD from a handler the handler will
+ *         need to have made sure to either re-enqueue the op or destroy it
+ *         since the caller will not touch the op anymore.
+ */
+typedef enum {
+        RD_KAFKA_OP_RES_PASS,    /* Not handled, pass to caller */
+        RD_KAFKA_OP_RES_HANDLED, /* Op was handled (through callbacks) */
+        RD_KAFKA_OP_RES_YIELD    /* Callback called yield */
+} rd_kafka_op_res_t;
+
+
+/**
+ * @brief Queue serve callback call type
+ */
+typedef enum {
+        RD_KAFKA_Q_CB_INVALID, /* dont use */
+        RD_KAFKA_Q_CB_CALLBACK,/* trigger callback based on op */
+        RD_KAFKA_Q_CB_RETURN,  /* return op rather than trigger callback
+                                * (if possible)*/
+        RD_KAFKA_Q_CB_FORCE_RETURN, /* return op, regardless of callback. */
+        RD_KAFKA_Q_CB_EVENT    /* like _Q_CB_RETURN but return event_t:ed op */
+} rd_kafka_q_cb_type_t;
+
+/**
+ * @brief Queue serve callback
+ * @remark See rd_kafka_op_res_t docs for return semantics.
+ */
+typedef rd_kafka_op_res_t
+(rd_kafka_q_serve_cb_t) (rd_kafka_t *rk,
+                         struct rd_kafka_q_s *rkq,
+                         struct rd_kafka_op_s *rko,
+                         rd_kafka_q_cb_type_t cb_type, void *opaque)
+        RD_WARN_UNUSED_RESULT;
+
+/**
+ * @brief Op callback type
+ */
+typedef rd_kafka_op_res_t (rd_kafka_op_cb_t) (rd_kafka_t *rk,
+                                              rd_kafka_q_t *rkq,
+                                              struct rd_kafka_op_s *rko)
+                RD_WARN_UNUSED_RESULT;
+
+
 #define RD_KAFKA_OP_TYPE_ASSERT(rko,type) \
 	rd_kafka_assert(NULL, (rko)->rko_type == (type) && # type)
 
@@ -127,6 +191,8 @@ struct rd_kafka_op_s {
 	rd_kafka_resp_err_t   rko_err;
 	int32_t               rko_len;    /* Depends on type, typically the
 					   * message length. */
+        rd_kafka_op_prio_t    rko_prio;   /* In-queue priority.
+                                           * Higher value means higher prio. */
 
 	shptr_rd_kafka_toppar_t *rko_rktp;
 
@@ -138,10 +204,20 @@ struct rd_kafka_op_s {
 	 * .q is refcounted. */
 	rd_kafka_replyq_t rko_replyq;
 
+        /* Original queue's op serve callback and opaque, if any.
+         * Mainly used for forwarded queues to use the original queue's
+         * serve function from the forwarded position. */
+        rd_kafka_q_serve_cb_t *rko_serve;
+        void *rko_serve_opaque;
+
 	rd_kafka_t     *rko_rk;
 
+#if ENABLE_DEVEL
+        const char *rko_source;  /**< Where op was created */
+#endif
+
         /* RD_KAFKA_OP_CB */
-        void          (*rko_op_cb) (rd_kafka_t *rk, struct rd_kafka_op_s *rko);
+        rd_kafka_op_cb_t *rko_op_cb;
 
 	union {
 		struct {
@@ -165,6 +241,7 @@ struct rd_kafka_op_s {
 			int silent_empty; /**< Fail silently if there are no
 					   *   offsets to commit. */
                         rd_ts_t ts_timeout;
+                        char *reason;
 		} offset_commit;
 
 		struct {
@@ -204,12 +281,8 @@ struct rd_kafka_op_s {
 			rd_kafka_buf_t *rkbuf;
 		} xbuf; /* XMIT_BUF and RECV_BUF */
 
-		struct {
-			rd_kafka_topic_t *rkt;
-			int  all_topics;
-			char reason[128];
-			struct rd_kafka_metadata *metadata;
-		} metadata;
+                /* RD_KAFKA_OP_METADATA */
+                rd_kafka_metadata_t *metadata;
 
 		struct {
 			shptr_rd_kafka_itopic_t *s_rkt;
@@ -237,6 +310,12 @@ struct rd_kafka_op_s {
 			int pause;
 			int flag;
 		} pause;
+
+                struct {
+                        char fac[64];
+                        int  level;
+                        char *str;
+                } log;
 	} rko_u;
 };
 
@@ -247,15 +326,23 @@ TAILQ_HEAD(rd_kafka_op_head_s, rd_kafka_op_s);
 
 const char *rd_kafka_op2str (rd_kafka_op_type_t type);
 void rd_kafka_op_destroy (rd_kafka_op_t *rko);
-rd_kafka_op_t *rd_kafka_op_new (rd_kafka_op_type_t type);
+rd_kafka_op_t *rd_kafka_op_new0 (const char *source, rd_kafka_op_type_t type);
+#if ENABLE_DEVEL
+#define _STRINGIFYX(A) #A
+#define _STRINGIFY(A) _STRINGIFYX(A)
+#define rd_kafka_op_new(type)                                   \
+        rd_kafka_op_new0(__FILE__ ":" _STRINGIFY(__LINE__), type)
+#else
+#define rd_kafka_op_new(type) rd_kafka_op_new0(NULL, type)
+#endif
 rd_kafka_op_t *rd_kafka_op_new_reply (rd_kafka_op_t *rko_orig,
-				      rd_kafka_resp_err_t err);
-
-
-
+                                      rd_kafka_resp_err_t err);
+rd_kafka_op_t *rd_kafka_op_new_cb (rd_kafka_t *rk,
+                                   rd_kafka_op_type_t type,
+                                   rd_kafka_op_cb_t *cb);
 int rd_kafka_op_reply (rd_kafka_op_t *rko, rd_kafka_resp_err_t err);
 
-
+#define rd_kafka_op_set_prio(rko,prio) ((rko)->rko_prio = prio)
 
 
 #define rd_kafka_op_err(rk,err,...) do {				\
@@ -277,13 +364,29 @@ rd_kafka_op_t *rd_kafka_op_req (rd_kafka_q_t *destq,
 rd_kafka_op_t *rd_kafka_op_req2 (rd_kafka_q_t *destq, rd_kafka_op_type_t type);
 rd_kafka_resp_err_t rd_kafka_op_err_destroy (rd_kafka_op_t *rko);
 
-void rd_kafka_op_call (rd_kafka_t *rk, rd_kafka_op_t *rko);
+rd_kafka_op_res_t rd_kafka_op_call (rd_kafka_t *rk,
+                                    rd_kafka_q_t *rkq, rd_kafka_op_t *rko)
+        RD_WARN_UNUSED_RESULT;
+
+rd_kafka_op_t *
+rd_kafka_op_new_fetch_msg (rd_kafka_msg_t **rkmp,
+                           rd_kafka_toppar_t *rktp,
+                           int32_t version,
+                           rd_kafka_buf_t *rkbuf,
+                           int64_t offset,
+                           size_t key_len, const void *key,
+                           size_t val_len, const void *val);
 
 void rd_kafka_op_throttle_time (struct rd_kafka_broker_s *rkb,
 				rd_kafka_q_t *rkq,
 				int throttle_time);
 
-int rd_kafka_op_handle_std (rd_kafka_t *rk, rd_kafka_op_t *rko);
+
+rd_kafka_op_res_t
+rd_kafka_op_handle (rd_kafka_t *rk, rd_kafka_q_t *rkq, rd_kafka_op_t *rko,
+                    rd_kafka_q_cb_type_t cb_type, void *opaque,
+                    rd_kafka_q_serve_cb_t *callback) RD_WARN_UNUSED_RESULT;
+
 
 extern rd_atomic32_t rd_kafka_op_cnt;
 
