@@ -29,6 +29,8 @@
 #pragma comment(lib, "ws2_32.lib")
 #endif
 
+#define __need_IOV_MAX
+
 #define _DARWIN_C_SOURCE  /* MSG_DONTWAIT */
 
 #include "rdkafka_int.h"
@@ -96,10 +98,7 @@ void rd_kafka_transport_close (rd_kafka_transport_t *rktrans) {
 	}
 #endif
 
-#if WITH_SASL
-	if (rktrans->rktrans_sasl.conn)
-		sasl_dispose(&rktrans->rktrans_sasl.conn);
-#endif
+        rd_kafka_sasl_close(rktrans);
 
 	if (rktrans->rktrans_recv_buf)
 		rd_kafka_buf_destroy(rktrans->rktrans_recv_buf);
@@ -115,8 +114,7 @@ void rd_kafka_transport_close (rd_kafka_transport_t *rktrans) {
 static const char *socket_strerror(int err) {
 #ifdef _MSC_VER
 	static RD_TLS char buf[256];
-	FormatMessageA(FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS, NULL,
-		       err, MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), (LPSTR)buf, sizeof(buf)-1, NULL);
+        rd_strerror_w32(err, buf, sizeof(buf));
 	return buf;
 #else
 	return rd_strerror(err);
@@ -126,104 +124,247 @@ static const char *socket_strerror(int err) {
 
 
 
-
-
-
-static ssize_t
-rd_kafka_transport_socket_sendmsg (rd_kafka_transport_t *rktrans,
-				   const struct msghdr *msg,
-				   char *errstr, size_t errstr_size) {
 #ifndef _MSC_VER
-	ssize_t r;
+/**
+ * @brief sendmsg() abstraction, converting a list of segments to iovecs.
+ * @remark should only be called if the number of segments is > 1.
+ */
+ssize_t rd_kafka_transport_socket_sendmsg (rd_kafka_transport_t *rktrans,
+                                           rd_slice_t *slice,
+                                           char *errstr, size_t errstr_size) {
+        struct iovec iov[IOV_MAX];
+        struct msghdr msg = { .msg_iov = iov };
+        size_t iovlen;
+        ssize_t r;
+
+        rd_slice_get_iov(slice, msg.msg_iov, &iovlen, IOV_MAX,
+                         /* FIXME: Measure the effects of this */
+                         rktrans->rktrans_sndbuf_size);
+        msg.msg_iovlen = (typeof(msg.msg_iovlen))iovlen;
 
 #ifdef sun
-	/* See recvmsg() comment. Setting it here to be safe. */
-	socket_errno = EAGAIN;
+        /* See recvmsg() comment. Setting it here to be safe. */
+        socket_errno = EAGAIN;
 #endif
-	r = sendmsg(rktrans->rktrans_s, msg, MSG_DONTWAIT
+
+        r = sendmsg(rktrans->rktrans_s, &msg, MSG_DONTWAIT
 #ifdef MSG_NOSIGNAL
-		    | MSG_NOSIGNAL
+                    | MSG_NOSIGNAL
 #endif
-		);
-	if (r == -1) {
-		if (socket_errno == EAGAIN)
-			return 0;
-		rd_snprintf(errstr, errstr_size, "%s", rd_strerror(errno));
-	}
-	return r;
+                );
+
+        if (r == -1) {
+                if (socket_errno == EAGAIN)
+                        return 0;
+                rd_snprintf(errstr, errstr_size, "%s", rd_strerror(errno));
+        }
+
+        /* Update buffer read position */
+        rd_slice_read(slice, NULL, (size_t)r);
+
+        return r;
+}
+#endif
+
+
+/**
+ * @brief Plain send() abstraction
+ */
+static ssize_t
+rd_kafka_transport_socket_send0 (rd_kafka_transport_t *rktrans,
+                                 rd_slice_t *slice,
+                                 char *errstr, size_t errstr_size) {
+        ssize_t sum = 0;
+        const void *p;
+        size_t rlen;
+
+        while ((rlen = rd_slice_peeker(slice, &p))) {
+                ssize_t r;
+
+                r = send(rktrans->rktrans_s, p,
+#ifdef _MSC_VER
+                         (int)rlen, (int)0
 #else
-	int i;
-	ssize_t sum = 0;
-
-	for (i = 0; i < msg->msg_iovlen; i++) {
-		ssize_t r;
-
-		r = send(rktrans->rktrans_s,
-			 msg->msg_iov[i].iov_base, (int) msg->msg_iov[i].iov_len, 0);
-		if (r == SOCKET_ERROR) {
-			if (sum > 0 || WSAGetLastError() == WSAEWOULDBLOCK)
-				return sum;
-			else {
-				rd_snprintf(errstr, errstr_size, "%s",
-					    socket_strerror(WSAGetLastError()));
-				return -1;
-			}
-		}
-
-		sum += r;
-		if ((size_t)r < msg->msg_iov[i].iov_len)
-			break;
-
-	}
-	return sum;
+                         rlen, 0
 #endif
+                );
+
+#ifdef _MSC_VER
+                if (unlikely(r == SOCKET_ERROR)) {
+                        if (sum > 0 || WSAGetLastError() == WSAEWOULDBLOCK)
+                                return sum;
+                        else {
+                                rd_snprintf(errstr, errstr_size, "%s",
+                                            socket_strerror(WSAGetLastError()));
+                                return -1;
+                        }
+                }
+#else
+                if (unlikely(r <= 0)) {
+                        if (r == 0 || errno == EAGAIN)
+                                return 0;
+                        rd_snprintf(errstr, errstr_size, "%s",
+                                    socket_strerror(socket_errno));
+                        return -1;
+                }
+#endif
+
+                /* Update buffer read position */
+                rd_slice_read(slice, NULL, (size_t)r);
+
+                sum += r;
+
+                /* FIXME: remove this and try again immediately and let
+                 *        the next write() call fail instead? */
+                if ((size_t)r < rlen)
+                        break;
+        }
+
+        return sum;
 }
 
 
+static ssize_t
+rd_kafka_transport_socket_send (rd_kafka_transport_t *rktrans,
+                                rd_slice_t *slice,
+                                char *errstr, size_t errstr_size) {
+#ifndef _MSC_VER
+        /* FIXME: Use sendmsg() with iovecs if there's more than one segment
+         * remaining, otherwise (or if platform does not have sendmsg)
+         * use plain send(). */
+        return rd_kafka_transport_socket_sendmsg(rktrans, slice,
+                                                 errstr, errstr_size);
+#endif
+        return rd_kafka_transport_socket_send0(rktrans, slice,
+                                               errstr, errstr_size);
+}
+
+
+
+#ifndef _MSC_VER
+/**
+ * @brief recvmsg() abstraction, converting a list of segments to iovecs.
+ * @remark should only be called if the number of segments is > 1.
+ */
 static ssize_t
 rd_kafka_transport_socket_recvmsg (rd_kafka_transport_t *rktrans,
-				   struct msghdr *msg,
-				   char *errstr, size_t errstr_size) {
-#ifndef _MSC_VER
-	ssize_t r;
+                                   rd_buf_t *rbuf,
+                                   char *errstr, size_t errstr_size) {
+        ssize_t r;
+        struct iovec iov[IOV_MAX];
+        struct msghdr msg = { .msg_iov = iov };
+        size_t iovlen;
+
+        rd_buf_get_write_iov(rbuf, msg.msg_iov, &iovlen, IOV_MAX,
+                             /* FIXME: Measure the effects of this */
+                             rktrans->rktrans_rcvbuf_size);
+        msg.msg_iovlen = (typeof(msg.msg_iovlen))iovlen;
+
 #ifdef sun
-	/* SunOS doesn't seem to set errno when recvmsg() fails
-	 * due to no data and MSG_DONTWAIT is set. */
-	socket_errno = EAGAIN;
+        /* SunOS doesn't seem to set errno when recvmsg() fails
+         * due to no data and MSG_DONTWAIT is set. */
+        socket_errno = EAGAIN;
 #endif
-	r = recvmsg(rktrans->rktrans_s, msg, MSG_DONTWAIT);
-	if (r == -1 && socket_errno == EAGAIN)
-		return 0;
-	else if (r == 0) {
-		/* Receive 0 after POLLIN event means connection closed. */
-		rd_snprintf(errstr, errstr_size, "Disconnected");
-		return -1;
-	} else if (r == -1)
-		rd_snprintf(errstr, errstr_size, "%s", rd_strerror(errno));
+        r = recvmsg(rktrans->rktrans_s, &msg, MSG_DONTWAIT);
+        if (unlikely(r <= 0)) {
+                if (r == -1 && socket_errno == EAGAIN)
+                        return 0;
+                else if (r == 0) {
+                        /* Receive 0 after POLLIN event means
+                         * connection closed. */
+                        rd_snprintf(errstr, errstr_size, "Disconnected");
+                        return -1;
+                } else if (r == -1) {
+                        rd_snprintf(errstr, errstr_size, "%s",
+                                    rd_strerror(errno));
+                        return -1;
+                }
+        }
 
-	return r;
-#else
-	ssize_t sum = 0;
-	int i;
+        /* Update buffer write position */
+        rd_buf_write(rbuf, NULL, (size_t)r);
 
-	for (i = 0; i < msg->msg_iovlen; i++) {
-		ssize_t r;
-
-		r = recv(rktrans->rktrans_s,
-			 msg->msg_iov[i].iov_base, (int) msg->msg_iov[i].iov_len, 0);
-		if (r == SOCKET_ERROR) {
-			if (WSAGetLastError() == WSAEWOULDBLOCK)
-				break;
-			rd_snprintf(errstr, errstr_size, "%s", socket_strerror(WSAGetLastError()));
-			return -1;
-		}
-		sum += r;
-		if ((size_t)r < msg->msg_iov[i].iov_len)
-			break;
-	}
-	return sum;
-#endif
+        return r;
 }
+#endif
+
+
+/**
+ * @brief Plain recv()
+ */
+static ssize_t
+rd_kafka_transport_socket_recv0 (rd_kafka_transport_t *rktrans,
+                                 rd_buf_t *rbuf,
+                                 char *errstr, size_t errstr_size) {
+        ssize_t sum = 0;
+        void *p;
+        size_t len;
+
+        while ((len = rd_buf_get_writable(rbuf, &p))) {
+                ssize_t r;
+
+                r = recv(rktrans->rktrans_s, p,
+#ifdef _MSC_VER
+                         (int)
+#endif
+                         len,
+                         0);
+
+#ifdef _MSC_VER
+                if (unlikely(r == SOCKET_ERROR)) {
+                        if (WSAGetLastError() == WSAEWOULDBLOCK)
+                                return sum;
+                        rd_snprintf(errstr, errstr_size, "%s",
+                                    socket_strerror(WSAGetLastError()));
+                        return -1;
+                }
+#else
+                if (unlikely(r <= 0)) {
+                        if (r == -1 && socket_errno == EAGAIN)
+                                return 0;
+                        else if (r == 0) {
+                                /* Receive 0 after POLLIN event means
+                                 * connection closed. */
+                                rd_snprintf(errstr, errstr_size,
+                                            "Disconnected");
+                                return -1;
+                        } else if (r == -1) {
+                                rd_snprintf(errstr, errstr_size, "%s",
+                                            rd_strerror(errno));
+                                return -1;
+                        }
+                }
+#endif
+
+                /* Update buffer write position */
+                rd_buf_write(rbuf, NULL, (size_t)r);
+
+                sum += r;
+
+                /* FIXME: remove this and try again immediately and let
+                 *        the next recv() call fail instead? */
+                if ((size_t)r < len)
+                        break;
+        }
+        return sum;
+}
+
+
+static ssize_t
+rd_kafka_transport_socket_recv (rd_kafka_transport_t *rktrans,
+                                rd_buf_t *buf,
+                                char *errstr, size_t errstr_size) {
+#ifndef _MSC_VER
+        /* FIXME: Use recvmsg() with iovecs if there's more than one segment
+         * remaining, otherwise (or if platform does not have sendmsg)
+         * use plain send(). */
+        return rd_kafka_transport_socket_recvmsg(rktrans, buf,
+                                                 errstr, errstr_size);
+#endif
+        return rd_kafka_transport_socket_recv0(rktrans, buf,
+                                               errstr, errstr_size);
+}
+
 
 
 
@@ -305,6 +446,7 @@ void rd_kafka_transport_ssl_term (void) {
 
 	CRYPTO_set_id_callback(NULL);
 	CRYPTO_set_locking_callback(NULL);
+        CRYPTO_cleanup_all_ex_data();
 
 	for (i = 0 ; i < rd_kafka_ssl_locks_cnt ; i++)
 		mtx_destroy(&rd_kafka_ssl_locks[i]);
@@ -371,6 +513,10 @@ rd_kafka_transport_ssl_io_update (rd_kafka_transport_t *rktrans, int ret,
 				    rd_strerror(errno));
 		return -1;
 
+        case SSL_ERROR_ZERO_RETURN:
+                rd_snprintf(errstr, errstr_size, "Disconnected");
+                return -1;
+
 	default:
 		rd_kafka_ssl_error(NULL, rktrans->rktrans_rkb,
 				   errstr, errstr_size);
@@ -381,21 +527,17 @@ rd_kafka_transport_ssl_io_update (rd_kafka_transport_t *rktrans, int ret,
 }
 
 static ssize_t
-rd_kafka_transport_ssl_sendmsg (rd_kafka_transport_t *rktrans,
-				const struct msghdr *msg,
-				char *errstr, size_t errstr_size) {
-	int i;
+rd_kafka_transport_ssl_send (rd_kafka_transport_t *rktrans,
+                             rd_slice_t *slice,
+                             char *errstr, size_t errstr_size) {
 	ssize_t sum = 0;
+        const void *p;
+        size_t rlen;
 
-	for (i = 0; i < (int)msg->msg_iovlen; i++) {
-		int r;
+        while ((rlen = rd_slice_peeker(slice, &p))) {
+                int r;
 
-		if (unlikely(msg->msg_iov[i].iov_len == 0))
-		    continue;
-
-		r = SSL_write(rktrans->rktrans_ssl, 
-			      msg->msg_iov[i].iov_base,
-			      (int) msg->msg_iov[i].iov_len);
+                r = SSL_write(rktrans->rktrans_ssl, p, (int)rlen);
 
 		if (unlikely(r <= 0)) {
 			if (rd_kafka_transport_ssl_io_update(rktrans, r,
@@ -406,27 +548,30 @@ rd_kafka_transport_ssl_sendmsg (rd_kafka_transport_t *rktrans,
 				return sum;
 		}
 
+                /* Update buffer read position */
+                rd_slice_read(slice, NULL, (size_t)r);
 
 		sum += r;
-		if ((size_t)r < msg->msg_iov[i].iov_len)
-			break;
+                 /* FIXME: remove this and try again immediately and let
+                  *        the next SSL_write() call fail instead? */
+                if ((size_t)r < rlen)
+                        break;
 
 	}
 	return sum;
 }
- 
+
 static ssize_t
-rd_kafka_transport_ssl_recvmsg (rd_kafka_transport_t *rktrans,
-				struct msghdr *msg,
-				char *errstr, size_t errstr_size) {
+rd_kafka_transport_ssl_recv (rd_kafka_transport_t *rktrans,
+                             rd_buf_t *rbuf, char *errstr, size_t errstr_size) {
 	ssize_t sum = 0;
-	int i;
-	
-	for (i = 0; i < (int)msg->msg_iovlen; i++) {
+        void *p;
+        size_t len;
+
+        while ((len = rd_buf_get_writable(rbuf, &p))) {
 		int r;
 
-		r = SSL_read(rktrans->rktrans_ssl,
-			     msg->msg_iov[i].iov_base, (int) msg->msg_iov[i].iov_len);
+                r = SSL_read(rktrans->rktrans_ssl, p, (int)len);
 
 		if (unlikely(r <= 0)) {
 			if (rd_kafka_transport_ssl_io_update(rktrans, r,
@@ -437,11 +582,18 @@ rd_kafka_transport_ssl_recvmsg (rd_kafka_transport_t *rktrans,
 				return sum;
 		}
 
-		VALGRIND_MAKE_MEM_DEFINED(msg->msg_iov[i].iov_base, r);
+                VALGRIND_MAKE_MEM_DEFINED(p, r);
+
+                /* Update buffer write position */
+                rd_buf_write(rbuf, NULL, (size_t)r);
 
 		sum += r;
-		if ((size_t)r < msg->msg_iov[i].iov_len)
-			break;
+
+                 /* FIXME: remove this and try again immediately and let
+                  *        the next SSL_read() call fail instead? */
+                if ((size_t)r < len)
+                        break;
+
 	}
 	return sum;
 
@@ -484,8 +636,10 @@ static int rd_kafka_transport_ssl_passwd_cb (char *buf, int size, int rwflag,
  */
 static int rd_kafka_transport_ssl_connect (rd_kafka_broker_t *rkb,
 					   rd_kafka_transport_t *rktrans,
-					   char *errstr, int errstr_size) {
+					   char *errstr, size_t errstr_size) {
 	int r;
+	char name[RD_KAFKA_NODENAME_SIZE];
+	char *t;
 
 	rktrans->rktrans_ssl = SSL_new(rkb->rkb_rk->rk_conf.ssl.ctx);
 	if (!rktrans->rktrans_ssl)
@@ -494,6 +648,18 @@ static int rd_kafka_transport_ssl_connect (rd_kafka_broker_t *rkb,
 	if (!SSL_set_fd(rktrans->rktrans_ssl, rktrans->rktrans_s))
 		goto fail;
 
+#if (OPENSSL_VERSION_NUMBER >= 0x0090806fL) && !defined(OPENSSL_NO_TLSEXT)
+	/* If non-numerical hostname, send it for SNI */
+	rd_snprintf(name, sizeof(name), "%s", rkb->rkb_nodename);
+	if ((t = strrchr(name, ':')))
+		*t = '\0';
+	if (!(/*ipv6*/(strchr(name, ':') &&
+		       strspn(name, "0123456789abcdefABCDEF:.[]%") == strlen(name)) ||
+	      /*ipv4*/strspn(name, "0123456789.") == strlen(name)) &&
+	    !SSL_set_tlsext_host_name(rktrans->rktrans_ssl, name))
+		goto fail;
+#endif
+
 	r = SSL_connect(rktrans->rktrans_ssl);
 	if (r == 1) {
 		/* Connected, highly unlikely since this is a
@@ -627,9 +793,15 @@ int rd_kafka_transport_ssl_ctx_init (rd_kafka_t *rk,
 	int r;
 	SSL_CTX *ctx;
 
+        if (errstr_size > 0)
+                errstr[0] = '\0';
+
 	ctx = SSL_CTX_new(SSLv23_client_method());
-	if (!ctx)
-		goto fail;
+        if (!ctx) {
+                rd_snprintf(errstr, errstr_size,
+                            "SSLv23_client_method() failed: ");
+                goto fail;
+        }
 
 #ifdef SSL_OP_NO_SSLv3
 	/* Disable SSLv3 (unsafe) */
@@ -647,9 +819,12 @@ int rd_kafka_transport_ssl_ctx_init (rd_kafka_t *rk,
 			     rk->rk_conf.ssl.cipher_suites);
 		if (!SSL_CTX_set_cipher_list(ctx,
 					     rk->rk_conf.ssl.cipher_suites)) {
-			rd_snprintf(errstr, errstr_size,
-				    "No recognized ciphers");
-			goto fail;
+                        /* Set a string that will prefix the
+                         * the OpenSSL error message (which is lousy)
+                         * to make it more meaningful. */
+                        rd_snprintf(errstr, errstr_size,
+                                    "ssl.cipher.suites failed: ");
+                        goto fail;
 		}
 	}
 
@@ -671,9 +846,19 @@ int rd_kafka_transport_ssl_ctx_init (rd_kafka_t *rk,
 						  rk->rk_conf.ssl.
 						  ca_location : NULL);
 
-		if (r != 1)
-			goto fail;
-	}
+                if (r != 1) {
+                        rd_snprintf(errstr, errstr_size,
+                                    "ssl.ca.location failed: ");
+                        goto fail;
+                }
+        } else {
+                /* Use default CA certificate paths: ignore failures. */
+                r = SSL_CTX_set_default_verify_paths(ctx);
+                if (r != 1)
+                        rd_kafka_dbg(rk, SECURITY, "SSL",
+                                     "SSL_CTX_set_default_verify_paths() "
+                                     "failed: ignoring");
+        }
 
 	if (rk->rk_conf.ssl.crl_location) {
 		rd_kafka_dbg(rk, SECURITY, "SSL",
@@ -684,8 +869,11 @@ int rd_kafka_transport_ssl_ctx_init (rd_kafka_t *rk,
 						  rk->rk_conf.ssl.crl_location,
 						  NULL);
 
-		if (r != 1)
-			goto fail;
+                if (r != 1) {
+                        rd_snprintf(errstr, errstr_size,
+                                    "ssl.crl.location failed: ");
+                        goto fail;
+                }
 
 
 		rd_kafka_dbg(rk, SECURITY, "SSL",
@@ -703,8 +891,11 @@ int rd_kafka_transport_ssl_ctx_init (rd_kafka_t *rk,
 		r = SSL_CTX_use_certificate_chain_file(ctx,
 						       rk->rk_conf.ssl.cert_location);
 
-		if (r != 1)
-			goto fail;
+                if (r != 1) {
+                        rd_snprintf(errstr, errstr_size,
+                                    "ssl.certificate.location failed: ");
+                        goto fail;
+                }
 	}
 
 	if (rk->rk_conf.ssl.key_location) {
@@ -715,8 +906,11 @@ int rd_kafka_transport_ssl_ctx_init (rd_kafka_t *rk,
 		r = SSL_CTX_use_PrivateKey_file(ctx,
 						rk->rk_conf.ssl.key_location,
 						SSL_FILETYPE_PEM);
-		if (r != 1)
-			goto fail;
+                if (r != 1) {
+                        rd_snprintf(errstr, errstr_size,
+                                    "ssl.key.location failed: ");
+                        goto fail;
+                }
 	}
 
 
@@ -726,7 +920,9 @@ int rd_kafka_transport_ssl_ctx_init (rd_kafka_t *rk,
 	return 0;
 
  fail:
-	rd_kafka_ssl_error(rk, NULL, errstr, errstr_size);
+        r = (int)strlen(errstr);
+        rd_kafka_ssl_error(rk, NULL, errstr+r,
+                           (int)errstr_size > r ? (int)errstr_size - r : 0);
 	SSL_CTX_free(ctx);
 
 	return -1;
@@ -737,33 +933,31 @@ int rd_kafka_transport_ssl_ctx_init (rd_kafka_t *rk,
 
 
 ssize_t
-rd_kafka_transport_sendmsg (rd_kafka_transport_t *rktrans,
-			    const struct msghdr *msg,
-			    char *errstr, size_t errstr_size) {
+rd_kafka_transport_send (rd_kafka_transport_t *rktrans,
+                         rd_slice_t *slice, char *errstr, size_t errstr_size) {
 
 #if WITH_SSL
-	if (rktrans->rktrans_ssl)
-		return rd_kafka_transport_ssl_sendmsg(rktrans, msg,
-						      errstr, errstr_size);
-	else
+        if (rktrans->rktrans_ssl)
+                return rd_kafka_transport_ssl_send(rktrans, slice,
+                                                   errstr, errstr_size);
+        else
 #endif
-		return rd_kafka_transport_socket_sendmsg(rktrans, msg,
-							 errstr, errstr_size);
+                return rd_kafka_transport_socket_send(rktrans, slice,
+                                                      errstr, errstr_size);
 }
 
 
 ssize_t
-rd_kafka_transport_recvmsg (rd_kafka_transport_t *rktrans,
-			    struct msghdr *msg,
-			    char *errstr, size_t errstr_size) {
+rd_kafka_transport_recv (rd_kafka_transport_t *rktrans, rd_buf_t *rbuf,
+                         char *errstr, size_t errstr_size) {
 #if WITH_SSL
 	if (rktrans->rktrans_ssl)
-		return rd_kafka_transport_ssl_recvmsg(rktrans, msg,
-						      errstr, errstr_size);
+                return rd_kafka_transport_ssl_recv(rktrans, rbuf,
+                                                   errstr, errstr_size);
 	else
 #endif
-		return rd_kafka_transport_socket_recvmsg(rktrans, msg,
-							 errstr, errstr_size);
+                return rd_kafka_transport_socket_recv(rktrans, rbuf,
+                                                      errstr, errstr_size);
 }
 
 
@@ -779,48 +973,46 @@ rd_kafka_transport_recvmsg (rd_kafka_transport_t *rktrans,
  *    0: still waiting for data (*rkbufp remains unset)
  *    1: data complete, (buffer returned in *rkbufp)
  */
-int rd_kafka_transport_framed_recvmsg (rd_kafka_transport_t *rktrans,
-				       rd_kafka_buf_t **rkbufp,
-				       char *errstr, size_t errstr_size) {
+int rd_kafka_transport_framed_recv (rd_kafka_transport_t *rktrans,
+                                    rd_kafka_buf_t **rkbufp,
+                                    char *errstr, size_t errstr_size) {
 	rd_kafka_buf_t *rkbuf = rktrans->rktrans_recv_buf;
 	ssize_t r;
 	const int log_decode_errors = 0;
 
 	/* States:
 	 *   !rktrans_recv_buf: initial state; set up buf to receive header.
-	 *    rkbuf_len == 0:   awaiting header
-	 *    rkbuf_len > 0:    awaiting payload
+	 *    rkbuf_totlen == 0:   awaiting header
+	 *    rkbuf_totlen > 0:    awaiting payload
 	 */
 
 	if (!rkbuf) {
-		rkbuf = rd_kafka_buf_new(NULL, 1, 4);
-		/* Point read buffer to main buffer. */
-		rkbuf->rkbuf_rbuf = rkbuf->rkbuf_buf;
-		rd_kafka_buf_push(rkbuf, rkbuf->rkbuf_buf, 4);
-		rkbuf->rkbuf_len = 0; /* read bytes is zero */
-
+                rkbuf = rd_kafka_buf_new(1, 4/*length field's length*/);
+                /* Set up buffer reader for the length field */
+                rd_buf_write_ensure(&rkbuf->rkbuf_buf, 4, 4);
 		rktrans->rktrans_recv_buf = rkbuf;
 	}
 
 
-	r = rd_kafka_transport_recvmsg(rktrans, &rkbuf->rkbuf_msg,
-				       errstr, errstr_size);
+        r = rd_kafka_transport_recv(rktrans, &rkbuf->rkbuf_buf,
+                                    errstr, errstr_size);
 	if (r == 0)
 		return 0;
 	else if (r == -1)
 		return -1;
 
-	rkbuf->rkbuf_wof += r;
-
-	if (rkbuf->rkbuf_len == 0) {
+	if (rkbuf->rkbuf_totlen == 0) {
 		/* Frame length not known yet. */
 		int32_t frame_len;
 
-		if (rkbuf->rkbuf_wof < sizeof(frame_len)) {
+		if (rd_buf_write_pos(&rkbuf->rkbuf_buf) < sizeof(frame_len)) {
 			/* Wait for entire frame header. */
 			return 0;
 		}
 
+                /* Initialize reader */
+                rd_slice_init(&rkbuf->rkbuf_reader, &rkbuf->rkbuf_buf, 0, 4);
+
 		/* Reader header: payload length */
 		rd_kafka_buf_read_i32(rkbuf, &frame_len);
 
@@ -832,7 +1024,7 @@ int rd_kafka_transport_framed_recvmsg (rd_kafka_transport_t *rktrans,
 			return -1;
 		}
 
-		rkbuf->rkbuf_len = frame_len;
+		rkbuf->rkbuf_totlen = 4 + frame_len;
 		if (frame_len == 0) {
 			/* Payload is empty, we're done. */
 			rktrans->rktrans_recv_buf = NULL;
@@ -842,14 +1034,14 @@ int rd_kafka_transport_framed_recvmsg (rd_kafka_transport_t *rktrans,
 
 		/* Allocate memory to hold entire frame payload in contigious
 		 * memory. */
-		rd_kafka_buf_alloc_recvbuf(rkbuf, frame_len);
+                rd_buf_write_ensure_contig(&rkbuf->rkbuf_buf, frame_len);
 
-		/* Try reading directly, there is probably more data available*/
-		return rd_kafka_transport_framed_recvmsg(rktrans, rkbufp,
-							 errstr, errstr_size);
+                /* Try reading directly, there is probably more data available*/
+                return rd_kafka_transport_framed_recv(rktrans, rkbufp,
+                                                      errstr, errstr_size);
 	}
 
-	if (rkbuf->rkbuf_wof == rkbuf->rkbuf_len) {
+	if (rd_buf_write_pos(&rkbuf->rkbuf_buf) == rkbuf->rkbuf_totlen) {
 		/* Payload is complete. */
 		rktrans->rktrans_recv_buf = NULL;
 		*rkbufp = rkbuf;
@@ -859,10 +1051,11 @@ int rd_kafka_transport_framed_recvmsg (rd_kafka_transport_t *rktrans,
 	/* Wait for more data */
 	return 0;
 
- err:
+ err_parse:
 	if (rkbuf)
 		rd_kafka_buf_destroy(rkbuf);
-	rd_snprintf(errstr, errstr_size, "Frame header parsing failed");
+        rd_snprintf(errstr, errstr_size, "Frame header parsing failed: %s",
+                    rd_kafka_err2str(rkbuf->rkbuf_err));
 	return -1;
 }
 
@@ -875,6 +1068,7 @@ int rd_kafka_transport_framed_recvmsg (rd_kafka_transport_t *rktrans,
  */
 static void rd_kafka_transport_connected (rd_kafka_transport_t *rktrans) {
 	rd_kafka_broker_t *rkb = rktrans->rktrans_rkb;
+        unsigned int slen;
 
         rd_rkb_dbg(rkb, BROKER, "CONNECT",
                    "Connected to %s",
@@ -907,6 +1101,34 @@ static void rd_kafka_transport_connected (rd_kafka_transport_t *rktrans) {
 				   socket_strerror(socket_errno));
 	}
 
+        /* Get send and receive buffer sizes to allow limiting
+         * the total number of bytes passed with iovecs to sendmsg()
+         * and recvmsg(). */
+        slen = sizeof(rktrans->rktrans_rcvbuf_size);
+        if (getsockopt(rktrans->rktrans_s, SOL_SOCKET, SO_RCVBUF,
+                       (void *)&rktrans->rktrans_rcvbuf_size,
+                       &slen) == SOCKET_ERROR) {
+                rd_rkb_log(rkb, LOG_WARNING, "RCVBUF",
+                           "Failed to get socket receive "
+                           "buffer size: %s: assuming 1MB",
+                           socket_strerror(socket_errno));
+                rktrans->rktrans_rcvbuf_size = 1024*1024;
+        } else if (rktrans->rktrans_rcvbuf_size < 1024 * 64)
+                rktrans->rktrans_rcvbuf_size = 1024*64; /* Use at least 64KB */
+
+        slen = sizeof(rktrans->rktrans_sndbuf_size);
+        if (getsockopt(rktrans->rktrans_s, SOL_SOCKET, SO_SNDBUF,
+                       (void *)&rktrans->rktrans_sndbuf_size,
+                       &slen) == SOCKET_ERROR) {
+                rd_rkb_log(rkb, LOG_WARNING, "RCVBUF",
+                           "Failed to get socket send "
+                           "buffer size: %s: assuming 1MB",
+                           socket_strerror(socket_errno));
+                rktrans->rktrans_sndbuf_size = 1024*1024;
+        } else if (rktrans->rktrans_sndbuf_size < 1024 * 64)
+                rktrans->rktrans_sndbuf_size = 1024*64; /* Use at least 64KB */
+
+
 #ifdef TCP_NODELAY
 	if (rkb->rkb_rk->rk_conf.socket_nagle_disable) {
 		int one = 1;
@@ -1021,9 +1243,6 @@ static void rd_kafka_transport_io_event (rd_kafka_transport_t *rktrans,
 		break;
 
 	case RD_KAFKA_BROKER_STATE_AUTH:
-#if WITH_SASL
-		rd_kafka_assert(NULL, rktrans->rktrans_sasl.conn != NULL);
-
 		/* SASL handshake */
 		if (rd_kafka_sasl_io_event(rktrans, events,
 					   errstr, sizeof(errstr)) == -1) {
@@ -1034,7 +1253,6 @@ static void rd_kafka_transport_io_event (rd_kafka_transport_t *rktrans,
 					     errstr);
 			return;
 		}
-#endif
 		break;
 
 	case RD_KAFKA_BROKER_STATE_APIVERSION_QUERY:
@@ -1102,7 +1320,7 @@ void rd_kafka_transport_io_serve (rd_kafka_transport_t *rktrans,
 rd_kafka_transport_t *rd_kafka_transport_connect (rd_kafka_broker_t *rkb,
 						  const rd_sockaddr_inx_t *sinx,
 						  char *errstr,
-						  int errstr_size) {
+						  size_t errstr_size) {
 	rd_kafka_transport_t *rktrans;
 	int s = -1;
 	int on = 1;
@@ -1143,28 +1361,13 @@ rd_kafka_transport_t *rd_kafka_transport_connect (rd_kafka_broker_t *rkb,
 #endif
 	}
 
-
-	/* Set the socket to non-blocking */
-#ifdef _MSC_VER
-	if (ioctlsocket(s, FIONBIO, &on) == SOCKET_ERROR) {
-		rd_snprintf(errstr, errstr_size,
-			    "Failed to set socket non-blocking: %s",
-			    socket_strerror(socket_errno));
-		goto err;
-	}
-#else
-	{
-		int fl = fcntl(s, F_GETFL, 0);
-		if (fl == -1 ||
-		    fcntl(s, F_SETFL, fl | O_NONBLOCK) == -1) {
-			rd_snprintf(errstr, errstr_size,
-				    "Failed to set socket non-blocking: %s",
-				    socket_strerror(socket_errno));
-			goto err;
-		}
-	}
-#endif
-
+        /* Set the socket to non-blocking */
+        if ((r = rd_fd_set_nonblocking(s))) {
+                rd_snprintf(errstr, errstr_size,
+                            "Failed to set socket non-blocking: %s",
+                            socket_strerror(r));
+                goto err;
+        }
 
 	rd_rkb_dbg(rkb, BROKER, "CONNECT", "Connecting to %s (%s) "
 		   "with socket %i",
@@ -1208,7 +1411,12 @@ rd_kafka_transport_t *rd_kafka_transport_connect (rd_kafka_broker_t *rkb,
 	rktrans = rd_calloc(1, sizeof(*rktrans));
 	rktrans->rktrans_rkb = rkb;
 	rktrans->rktrans_s = s;
-	rktrans->rktrans_pfd.fd = s;
+	rktrans->rktrans_pfd[rktrans->rktrans_pfd_cnt++].fd = s;
+        if (rkb->rkb_wakeup_fd[0] != -1) {
+                rktrans->rktrans_pfd[rktrans->rktrans_pfd_cnt].events = POLLIN;
+                rktrans->rktrans_pfd[rktrans->rktrans_pfd_cnt++].fd = rkb->rkb_wakeup_fd[0];
+        }
+
 
 	/* Poll writability to trigger on connection success/failure. */
 	rd_kafka_transport_poll_set(rktrans, POLLOUT);
@@ -1225,26 +1433,22 @@ rd_kafka_transport_t *rd_kafka_transport_connect (rd_kafka_broker_t *rkb,
 
 
 void rd_kafka_transport_poll_set(rd_kafka_transport_t *rktrans, int event) {
-	rktrans->rktrans_pfd.events |= event;
+	rktrans->rktrans_pfd[0].events |= event;
 }
 
 void rd_kafka_transport_poll_clear(rd_kafka_transport_t *rktrans, int event) {
-	rktrans->rktrans_pfd.events &= ~event;
+	rktrans->rktrans_pfd[0].events &= ~event;
 }
 
 
 int rd_kafka_transport_poll(rd_kafka_transport_t *rktrans, int tmout) {
+        int r;
 #ifndef _MSC_VER
-	int r;
-
-	r = poll(&rktrans->rktrans_pfd, 1, tmout);
+	r = poll(rktrans->rktrans_pfd, rktrans->rktrans_pfd_cnt, tmout);
 	if (r <= 0)
 		return r;
-	return rktrans->rktrans_pfd.revents;
 #else
-	int r;
-
-	r = WSAPoll(&rktrans->rktrans_pfd, 1, tmout);
+	r = WSAPoll(rktrans->rktrans_pfd, rktrans->rktrans_pfd_cnt, tmout);
 	if (r == 0) {
 		/* Workaround for broken WSAPoll() while connecting:
 		 * failed connection attempts are not indicated at all by WSAPoll()
@@ -1271,8 +1475,19 @@ int rd_kafka_transport_poll(rd_kafka_transport_t *rktrans, int tmout) {
 			return 0;
 	} else if (r == SOCKET_ERROR)
 		return -1;
-	return rktrans->rktrans_pfd.revents;
 #endif
+        rd_atomic64_add(&rktrans->rktrans_rkb->rkb_c.wakeups, 1);
+
+        if (rktrans->rktrans_pfd[1].revents & POLLIN) {
+                /* Read wake-up fd data and throw away, just used for wake-ups*/
+                char buf[512];
+                if (rd_read((int)rktrans->rktrans_pfd[1].fd,
+                            buf, sizeof(buf)) == -1) {
+                        /* Ignore warning */
+                }
+        }
+
+        return rktrans->rktrans_pfd[0].revents;
 }
 
 
