@@ -39,23 +39,30 @@ void rd_list_dump (const char *what, const rd_list_t *rl) {
                        rl->rl_elems[i], &rl->rl_elems[i]);
 }
 
-static void rd_list_grow (rd_list_t *rl, int add_size) {
-	rd_assert(!(rl->rl_flags & RD_LIST_F_FIXED_SIZE));
-        rl->rl_size += add_size;
+void rd_list_grow (rd_list_t *rl, size_t size) {
+        rd_assert(!(rl->rl_flags & RD_LIST_F_FIXED_SIZE));
+        rl->rl_size += (int)size;
+        if (unlikely(rl->rl_size == 0))
+                return; /* avoid zero allocations */
         rl->rl_elems = rd_realloc(rl->rl_elems,
                                   sizeof(*rl->rl_elems) * rl->rl_size);
 }
 
-void rd_list_init (rd_list_t *rl, int initial_size) {
+rd_list_t *
+rd_list_init (rd_list_t *rl, int initial_size, void (*free_cb) (void *)) {
         memset(rl, 0, sizeof(*rl));
 
 	if (initial_size > 0)
 		rd_list_grow(rl, initial_size);
+
+        rl->rl_free_cb = free_cb;
+
+        return rl;
 }
 
-rd_list_t *rd_list_new (int initial_size) {
+rd_list_t *rd_list_new (int initial_size, void (*free_cb) (void *)) {
 	rd_list_t *rl = malloc(sizeof(*rl));
-	rd_list_init(rl, initial_size);
+	rd_list_init(rl, initial_size, free_cb);
 	rl->rl_flags |= RD_LIST_F_ALLOCATED;
 	return rl;
 }
@@ -88,8 +95,9 @@ void rd_list_prealloc_elems (rd_list_t *rl, size_t elemsize, size_t size) {
 }
 
 
-void rd_list_set_free_cb (rd_list_t *rl, void (*free_cb) (void *)) {
-	rl->rl_free_cb = free_cb;
+void rd_list_free_cb (rd_list_t *rl, void *ptr) {
+        if (rl->rl_free_cb && ptr)
+                rl->rl_free_cb(ptr);
 }
 
 
@@ -110,7 +118,6 @@ static void rd_list_remove0 (rd_list_t *rl, int idx) {
                         &rl->rl_elems[idx+1],
                         sizeof(*rl->rl_elems) * (rl->rl_cnt - (idx+1)));
         rl->rl_cnt--;
-	rl->rl_flags &= ~RD_LIST_F_SORTED;
 }
 
 void *rd_list_remove (rd_list_t *rl, void *match_elem) {
@@ -176,17 +183,14 @@ void rd_list_clear (rd_list_t *rl) {
 }
 
 
-void rd_list_destroy (rd_list_t *rl, void (*free_cb) (void *)) {
-
-	if (!free_cb)
-		free_cb = rl->rl_free_cb;
+void rd_list_destroy (rd_list_t *rl) {
 
         if (rl->rl_elems) {
                 int i;
-                if (free_cb) {
+                if (rl->rl_free_cb) {
                         for (i = 0 ; i < rl->rl_cnt ; i++)
                                 if (rl->rl_elems[i])
-                                        free_cb(rl->rl_elems[i]);
+                                        rl->rl_free_cb(rl->rl_elems[i]);
                 }
 
 		rd_free(rl->rl_elems);
@@ -242,3 +246,68 @@ int rd_list_cmp (const rd_list_t *a, rd_list_t *b,
 
 	return 0;
 }
+
+
+/**
+ * @brief Simple element pointer comparator
+ */
+int rd_list_cmp_ptr (const void *a, const void *b) {
+        if (a < b)
+                return -1;
+        else if (a > b)
+                return 1;
+        return 0;
+}
+
+
+void rd_list_apply (rd_list_t *rl,
+                    int (*cb) (void *elem, void *opaque), void *opaque) {
+        void *elem;
+        int i;
+
+        RD_LIST_FOREACH(elem, rl, i) {
+                if (!cb(elem, opaque)) {
+                        rd_list_remove0(rl, i);
+                        i--;
+                }
+        }
+
+        return;
+}
+
+
+/**
+ * @brief Default element copier that simply assigns the original pointer.
+ */
+static void *rd_list_nocopy_ptr (const void *elem, void *opaque) {
+        return (void *)elem;
+}
+
+
+rd_list_t *rd_list_copy (const rd_list_t *src,
+                         void *(*copy_cb) (const void *elem, void *opaque),
+                         void *opaque) {
+        rd_list_t *dst;
+
+        dst = rd_list_new(src->rl_cnt, src->rl_free_cb);
+
+        rd_list_copy_to(dst, src, copy_cb, opaque);
+        return dst;
+}
+
+
+void rd_list_copy_to (rd_list_t *dst, const rd_list_t *src,
+                      void *(*copy_cb) (const void *elem, void *opaque),
+                      void *opaque) {
+        void *elem;
+        int i;
+
+        if (!copy_cb)
+                copy_cb = rd_list_nocopy_ptr;
+
+        RD_LIST_FOREACH(elem, src, i) {
+                void *celem = copy_cb(elem, opaque);
+                if (celem)
+                        rd_list_add(dst, celem);
+        }
+}
