@@ -36,6 +36,9 @@
  * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  */
 
+#pragma GCC diagnostic push
+#pragma GCC diagnostic ignored "-Wcast-align"
+
 #ifndef SG
 #define SG /* Scatter-Gather / iovec support in Snappy */
 #endif
@@ -1481,29 +1484,23 @@ out:
 #ifdef SG
 
 int rd_kafka_snappy_compress_iov(struct snappy_env *env,
-			struct iovec *iov_in,
-			int iov_in_len,
-			size_t input_length,
-			struct iovec *iov_out,
-			int *iov_out_len,
-			size_t *compressed_length)
-{
-	struct source reader = {
-		.iov = iov_in,
-		.iovlen = iov_in_len,
-		.total = input_length
-	};
-	struct sink writer = {
-		.iov = iov_out,
-		.iovlen = *iov_out_len,
-	};
-	int err = sn_compress(env, &reader, &writer);
+                                 const struct iovec *iov_in, size_t iov_in_cnt,
+                                 size_t input_length,
+                                 struct iovec *iov_out) {
+        struct source reader = {
+                .iov = (struct iovec *)iov_in,
+                .iovlen = (int)iov_in_cnt,
+                .total = input_length
+        };
+        struct sink writer = {
+                .iov = iov_out,
+                .iovlen = 1
+        };
+        int err = sn_compress(env, &reader, &writer);
 
-	*iov_out_len = writer.curvec + 1;
+        iov_out->iov_len = writer.written;
 
-	/* Compute how many bytes were added */
-	*compressed_length = writer.written;
-	return err;
+        return err;
 }
 EXPORT_SYMBOL(rd_kafka_snappy_compress_iov);
 
@@ -1537,10 +1534,9 @@ int rd_kafka_snappy_compress(struct snappy_env *env,
 		.iov_base = compressed,
 		.iov_len = 0xffffffff,
 	};
-	int out = 1;
-	return rd_kafka_snappy_compress_iov(env, 
-				   &iov_in, 1, input_length, 
-				   &iov_out, &out, compressed_length);
+        return rd_kafka_snappy_compress_iov(env,
+                                            &iov_in, 1, input_length,
+                                            &iov_out);
 }
 EXPORT_SYMBOL(rd_kafka_snappy_compress);
 
@@ -1581,6 +1577,122 @@ int rd_kafka_snappy_uncompress(const char *compressed, size_t n, char *uncompres
 }
 EXPORT_SYMBOL(rd_kafka_snappy_uncompress);
 
+
+/**
+ * @brief Decompress Snappy message with Snappy-java framing.
+ *
+ * @returns a malloced buffer with the uncompressed data, or NULL on failure.
+ */
+char *rd_kafka_snappy_java_uncompress (const char *inbuf, size_t inlen,
+                                       size_t *outlenp,
+                                       char *errstr, size_t errstr_size) {
+        int pass;
+        char *outbuf = NULL;
+
+        /**
+         * Traverse all chunks in two passes:
+         *  pass 1: calculate total uncompressed length
+         *  pass 2: uncompress
+         *
+         * Each chunk is prefixed with 4: length */
+
+        for (pass = 1 ; pass <= 2 ; pass++) {
+                ssize_t of = 0;  /* inbuf offset */
+                ssize_t uof = 0; /* outbuf offset */
+
+                while (of + 4 <= (ssize_t)inlen) {
+                        uint32_t clen; /* compressed length */
+                        size_t ulen; /* uncompressed length */
+                        int r;
+
+                        memcpy(&clen, inbuf+of, 4);
+                        clen = be32toh(clen);
+                        of += 4;
+
+                        if (unlikely(clen > inlen - of)) {
+                                rd_snprintf(errstr, errstr_size,
+                                            "Invalid snappy-java chunk length "
+                                            "%"PRId32" > %"PRIdsz
+                                            " available bytes",
+                                            clen, (ssize_t)inlen - of);
+                                return NULL;
+                        }
+
+                        /* Acquire uncompressed length */
+                        if (unlikely(!rd_kafka_snappy_uncompressed_length(
+                                             inbuf+of, clen, &ulen))) {
+                                rd_snprintf(errstr, errstr_size,
+                                            "Failed to get length of "
+                                            "(snappy-java framed) Snappy "
+                                            "compressed payload "
+                                            "(clen %"PRId32")",
+                                            clen);
+                                return NULL;
+                        }
+
+                        if (pass == 1) {
+                                /* pass 1: calculate total length */
+                                of  += clen;
+                                uof += ulen;
+                                continue;
+                        }
+
+                        /* pass 2: Uncompress to outbuf */
+                        if (unlikely((r = rd_kafka_snappy_uncompress(
+                                              inbuf+of, clen, outbuf+uof)))) {
+                                rd_snprintf(errstr, errstr_size,
+                                            "Failed to decompress Snappy-java "
+                                            "framed payload of size %"PRId32
+                                            ": %s",
+                                            clen,
+                                            rd_strerror(-r/*negative errno*/));
+                                rd_free(outbuf);
+                                return NULL;
+                        }
+
+                        of  += clen;
+                        uof += ulen;
+                }
+
+                if (unlikely(of != (ssize_t)inlen)) {
+                        rd_snprintf(errstr, errstr_size,
+                                    "%"PRIusz" trailing bytes in Snappy-java "
+                                    "framed compressed data",
+                                    inlen - of);
+                        if (outbuf)
+                                rd_free(outbuf);
+                        return NULL;
+                }
+
+                if (pass == 1) {
+                        if (uof <= 0) {
+                                rd_snprintf(errstr, errstr_size,
+                                            "Empty Snappy-java framed data");
+                                return NULL;
+                        }
+
+                        /* Allocate memory for uncompressed data */
+                        outbuf = rd_malloc(uof);
+                        if (unlikely(!outbuf)) {
+                                rd_snprintf(errstr, errstr_size,
+                                           "Failed to allocate memory "
+                                            "(%"PRIdsz") for "
+                                            "uncompressed Snappy data: %s",
+                                            uof, rd_strerror(errno));
+                                return NULL;
+                        }
+
+                } else {
+                        /* pass 2 */
+                        *outlenp = uof;
+                }
+        }
+
+        return outbuf;
+}
+
+
+
 #else
 /**
  * rd_kafka_snappy_compress - Compress a buffer using the snappy compressor.
@@ -1718,3 +1830,5 @@ void rd_kafka_snappy_free_env(struct snappy_env *env)
 	clear_env(env);
 }
 EXPORT_SYMBOL(rd_kafka_snappy_free_env);
+
+#pragma GCC diagnostic pop /* -Wcast-align ignore */
