@@ -40,7 +40,7 @@
 int main_0001_multiobj (int argc, char **argv) {
 	int partition = RD_KAFKA_PARTITION_UA; /* random */
 	int i;
-	const int NUM_ITER = 10;
+	const int NUM_ITER = 1;
         const char *topic = NULL;
 
 	TEST_SAY("Creating and destroying %i kafka instances\n", NUM_ITER);
@@ -52,20 +52,24 @@ int main_0001_multiobj (int argc, char **argv) {
 		rd_kafka_conf_t *conf;
 		rd_kafka_topic_conf_t *topic_conf;
 		char msg[128];
-                test_timing_t t_destroy;
+                test_timing_t t_full, t_destroy;
 
 		test_conf_init(&conf, &topic_conf, 30);
 
                 if (!topic)
                         topic = test_mk_topic_name("0001", 0);
 
+                /* Speed up produce to cut down on cycle time */
+                test_conf_set(conf, "queue.buffering.max.ms", "10");
+
+                TIMING_START(&t_full, "full create-produce-destroy cycle");
 		rk = test_create_handle(RD_KAFKA_PRODUCER, conf);
 
 		rkt = rd_kafka_topic_new(rk, topic, topic_conf);
 		if (!rkt)
 			TEST_FAIL("Failed to create topic for "
 				  "rdkafka instance #%i: %s\n",
-				  i, rd_kafka_err2str(rd_kafka_errno2err(errno)));
+				  i, rd_kafka_err2str(rd_kafka_last_error()));
 
 		rd_snprintf(msg, sizeof(msg), "%s test message for iteration #%i",
 			 argv[0], i);
@@ -84,6 +88,8 @@ int main_0001_multiobj (int argc, char **argv) {
                 TIMING_START(&t_destroy, "rd_kafka_destroy()");
 		rd_kafka_destroy(rk);
                 TIMING_STOP(&t_destroy);
+
+                TIMING_STOP(&t_full);
 	}
 
 	return 0;
