@@ -97,19 +97,81 @@ static void conf_cmp (const char *desc,
 		if (strcmp(a[i], b[i]))
 			TEST_FAIL("%s conf mismatch: %s != %s",
 				  desc, a[i], b[i]);
-		else if (strcmp(a[i+1], b[i+1]))
-			TEST_FAIL("%s conf value mismatch for %s: %s != %s",
-				  desc, a[i], a[i+1], b[i+1]);
+		else if (strcmp(a[i+1], b[i+1])) {
+                        /* The default_topic_conf will be auto-created
+                         * when global->topic fallthru is used, so its
+                         * value will not match here. */
+                        if (!strcmp(a[i], "default_topic_conf"))
+                                continue;
+                        TEST_FAIL("%s conf value mismatch for %s: %s != %s",
+                                  desc, a[i], a[i+1], b[i+1]);
+                }
 	}
 }
 
 
+/**
+ * @brief Not called, just used for config
+ */
+static int on_new_call_cnt;
+static rd_kafka_resp_err_t my_on_new (rd_kafka_t *rk,
+                                      const rd_kafka_conf_t *conf,
+                                      void *ic_opaque,
+                                      char *errstr, size_t errstr_size) {
+        TEST_SAY("%s: on_new() called\n", rd_kafka_name(rk));
+        on_new_call_cnt++;
+        return RD_KAFKA_RESP_ERR_NO_ERROR;
+}
+
+
+
+/**
+ * @brief When rd_kafka_new() succeeds it takes ownership of the config object,
+ *        but when it fails the config object remains in application custody.
+ *        These tests makes sure that's the case (preferably run with valgrind)
+ */
+static void do_test_kafka_new_failures (void) {
+        rd_kafka_conf_t *conf;
+        rd_kafka_t *rk;
+        char errstr[512];
+
+        conf = rd_kafka_conf_new();
+
+        rk = rd_kafka_new(RD_KAFKA_PRODUCER, conf, errstr, sizeof(errstr));
+        TEST_ASSERT(rk, "kafka_new() failed: %s", errstr);
+        rd_kafka_destroy(rk);
+
+        /* Set an erroneous configuration value that is not checked
+         * by conf_set() but by rd_kafka_new() */
+        conf = rd_kafka_conf_new();
+        if (rd_kafka_conf_set(conf, "partition.assignment.strategy",
+                              "range,thiswillfail", errstr, sizeof(errstr)) !=
+            RD_KAFKA_CONF_OK)
+                TEST_FAIL("%s", errstr);
+
+        rk = rd_kafka_new(RD_KAFKA_PRODUCER, conf, errstr, sizeof(errstr));
+        TEST_ASSERT(!rk, "kafka_new() should have failed");
+
+        /* config object should still belong to us,
+         * correct the erroneous config and try again. */
+        if (rd_kafka_conf_set(conf, "partition.assignment.strategy", NULL,
+                              errstr, sizeof(errstr)) !=
+            RD_KAFKA_CONF_OK)
+                TEST_FAIL("%s", errstr);
+
+        rk = rd_kafka_new(RD_KAFKA_PRODUCER, conf, errstr, sizeof(errstr));
+        TEST_ASSERT(rk, "kafka_new() failed: %s", errstr);
+        rd_kafka_destroy(rk);
+}
+
+
 int main_0004_conf (int argc, char **argv) {
 	rd_kafka_t *rk;
 	rd_kafka_topic_t *rkt;
 	rd_kafka_conf_t *ignore_conf, *conf, *conf2;
 	rd_kafka_topic_conf_t *ignore_topic_conf, *tconf, *tconf2;
 	char errstr[512];
+        rd_kafka_resp_err_t err;
 	const char **arr_orig, **arr_dup;
 	size_t cnt_orig, cnt_dup;
 	int i;
@@ -117,8 +179,9 @@ int main_0004_conf (int argc, char **argv) {
 	static const char *gconfs[] = {
 		"message.max.bytes", "12345", /* int property */
 		"client.id", "my id", /* string property */
-		"debug", "topic,metadata", /* S2F property */
+		"debug", "topic,metadata,interceptor", /* S2F property */
 		"topic.blacklist", "__.*", /* #778 */
+                "auto.offset.reset", "earliest", /* Global->Topic fallthru */
 #if WITH_ZLIB
 		"compression.codec", "gzip", /* S2I property */
 #endif
@@ -141,15 +204,21 @@ int main_0004_conf (int argc, char **argv) {
 	/* Set up a global config object */
 	conf = rd_kafka_conf_new();
 
+        for (i = 0 ; gconfs[i] ; i += 2) {
+                if (rd_kafka_conf_set(conf, gconfs[i], gconfs[i+1],
+                                      errstr, sizeof(errstr)) !=
+                    RD_KAFKA_CONF_OK)
+                        TEST_FAIL("%s\n", errstr);
+        }
+
 	rd_kafka_conf_set_dr_cb(conf, dr_cb);
 	rd_kafka_conf_set_error_cb(conf, error_cb);
-
-	for (i = 0 ; gconfs[i] ; i += 2) {
-		if (rd_kafka_conf_set(conf, gconfs[i], gconfs[i+1],
-				      errstr, sizeof(errstr)) !=
-		    RD_KAFKA_CONF_OK)
-			TEST_FAIL("%s\n", errstr);
-	}
+        /* interceptor configs are not exposed as strings or in dumps
+         * so the dump verification step will not cover them, but valgrind
+         * will help track down memory leaks/use-after-free etc. */
+        err = rd_kafka_conf_interceptor_add_on_new(conf, "testic",
+                                                   my_on_new, NULL);
+        TEST_ASSERT(!err, "add_on_new() failed: %s", rd_kafka_err2str(err));
 
 	/* Set up a topic config object */
 	tconf = rd_kafka_topic_conf_new();
@@ -195,7 +264,12 @@ int main_0004_conf (int argc, char **argv) {
 	 */
 
 	/* original */
+        TEST_ASSERT(on_new_call_cnt == 0, "expected 0 on_new call, not %d",
+                    on_new_call_cnt);
+        on_new_call_cnt = 0;
 	rk = test_create_handle(RD_KAFKA_PRODUCER, conf);
+        TEST_ASSERT(on_new_call_cnt == 1, "expected 1 on_new call, not %d",
+                on_new_call_cnt);
 
 	rkt = rd_kafka_topic_new(rk, topic, tconf);
 	if (!rkt)
@@ -206,13 +280,15 @@ int main_0004_conf (int argc, char **argv) {
 	rd_kafka_destroy(rk);
 
 	/* copied */
+        on_new_call_cnt = 0; /* interceptors are not copied. */
 	rk = test_create_handle(RD_KAFKA_PRODUCER, conf2);
+        TEST_ASSERT(on_new_call_cnt == 0, "expected 0 on_new call, not %d",
+                on_new_call_cnt);
 
 	rkt = rd_kafka_topic_new(rk, topic, tconf2);
 	if (!rkt)
 		TEST_FAIL("Failed to create topic: %s\n",
 			  rd_strerror(errno));
-
 	rd_kafka_topic_destroy(rkt);
 	rd_kafka_destroy(rk);
 
@@ -268,6 +344,7 @@ int main_0004_conf (int argc, char **argv) {
 			{ "request.required.acks", "393", "393" },
 			{ "request.required.acks", "bad", NULL },
 			{ "request.required.acks", "all", "-1" },
+                        { "request.required.acks", "all", "-1", 1/*fallthru*/ },
 			{ "acks", "0", "0" }, /* alias test */
 #if WITH_SASL
 			{ "sasl.mechanisms", "GSSAPI", "GSSAPI", 1 },
@@ -287,8 +364,9 @@ int main_0004_conf (int argc, char **argv) {
 			size_t destsz;
 			rd_kafka_conf_res_t res;
 
-			TEST_SAY("  Set: %s=%s expect %s\n",
-				 props[i].prop, props[i].val, props[i].exp);
+			TEST_SAY("  Set: %s=%s expect %s (%s)\n",
+				 props[i].prop, props[i].val, props[i].exp,
+                                 props[i].is_global ? "global":"topic");
 
 
 			/* Set value */
@@ -325,7 +403,8 @@ int main_0004_conf (int argc, char **argv) {
 							      props[i].prop,
 							      dest, &destsz);
 			TEST_ASSERT(res == RD_KAFKA_CONF_OK,
-				    ".._conf_get() returned %d", res);
+				    ".._conf_get(%s) returned %d",
+                                    props[i].prop, res);
 
 			TEST_ASSERT(!strcmp(props[i].exp, dest),
 				    "Expected \"%s\", got \"%s\"",
@@ -335,5 +414,7 @@ int main_0004_conf (int argc, char **argv) {
 		rd_kafka_conf_destroy(conf);
 	}
 
+        do_test_kafka_new_failures();
+
 	return 0;
 }
